import unittest
import spydrnet as sdn


class TestGetDefinitions(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.netlist = sdn.Netlist()

        leaf_library = cls.netlist.create_library()
        leaf_library.name = 'primitives'

        library = cls.netlist.create_library()
        library.name = 'work'

        leaf_def = leaf_library.create_definition()
        leaf_def.name = 'leaf'
        leaf_port = leaf_def.create_port()
        leaf_port.name = 'I'
        leaf_port.create_pins(1)

        bottom_def = library.create_definition()
        bottom_def.name = 'bottom'
        bottom_port = bottom_def.create_port()
        bottom_port.name = 'I'
        bottom_port.create_pins(1)
        leaf_inst = bottom_def.create_child()
        leaf_inst.reference = leaf_def
        bottom_cable = bottom_def.create_cable()
        bottom_cable.name = 'bottom_cable'
        bottom_wire = bottom_cable.create_wire()
        bottom_wire.connect_pin(bottom_port.pins[0])
        bottom_wire.connect_pin(leaf_inst.pins[leaf_port.pins[0]])
        cls.leaf_inst = leaf_inst

        bottom_floating_wire = bottom_cable.create_wire()

        middle_def = library.create_definition()
        middle_def.name = 'middle'
        middle_port = middle_def.create_port()
        middle_port.name = "I"
        middle_port.create_pin()
        bottom_inst = middle_def.create_child()
        bottom_inst.name = 'bottom'
        bottom_inst.reference = bottom_def
        middle_cable = middle_def.create_cable()
        middle_cable.name = "middle_cable"
        middle_wire = middle_cable.create_wire()
        middle_wire.connect_pin(middle_port.pins[0])
        middle_wire.connect_pin(bottom_inst.pins[bottom_port.pins[0]])
        cls.bottom_inst = bottom_inst

        middle_floating_wire = middle_cable.create_wire()

        top_def = library.create_definition()
        top_def.name = 'top'
        top_port = top_def.create_port()
        top_port.name = "I"
        top_port.create_pin()
        middle_inst = top_def.create_child()
        middle_inst.name = 'middle'
        middle_inst.reference = middle_def
        top_cable = top_def.create_cable()
        top_cable.name = "top_cable"
        top_wire = top_cable.create_wire()
        top_wire.connect_pin(top_port.pins[0])
        top_wire.connect_pin(middle_inst.pins[middle_port.pins[0]])

        top_floating_wire = top_cable.create_wire()

        top_instance = sdn.Instance()
        top_instance.name = 'top'
        top_instance.reference = top_def
        cls.netlist.top_instance = top_instance

    def test_parameter_checking(self):
        library = sdn.Library()
        definition = library.create_definition()
        definition.name = "MY_DEF"
        self.assertRaises(TypeError, sdn.get_definitions, library, "MY_DEF", patterns="MY_DEF")
        self.assertRaises(TypeError, sdn.get_definitions, library, "MY_DEF", unsupported_keyword=None)
        self.assertRaises(TypeError, sdn.get_definitions, library, "MY_DEF", selection="BOTH")
        self.assertRaises(TypeError, sdn.get_definitions, None, "MY_DEF")
        self.assertRaises(TypeError, sdn.get_definitions, [None, library], "MY_DEF")

    def test_collection(self):
        library = sdn.Library()
        definition = library.create_definition()
        definition.name = "MY_DEF"
        ports = list(sdn.get_definitions([library, library]))
        self.assertEqual(len(ports), 1)

    def test_get_definition_of_instances_outside(self):
        definition_query = list(sdn.get_definitions(self.netlist.libraries[0].definitions[0].references,
                                                    selection="OUTSIDE"))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.bottom_inst.reference)

    def test_get_definition_of_instances_inside(self):
        definition_query = list(sdn.get_definitions(self.netlist.libraries[0].definitions[0].references,
                                               selection="INSIDE"))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.leaf_inst.reference)

    def test_get_definition_of_instances_recursive_down(self):
        definition_query = set(sdn.get_definitions(self.netlist.top_instance, recursive=True))
        self.assertTrue(len(definition_query) == 4)

    def test_get_definition_of_instances_recursive_up(self):
        definition_query = list(sdn.get_definitions(self.leaf_inst, selection="OUTSIDE", recursive=True))
        self.assertTrue(len(definition_query) == 3 and self.leaf_inst.reference not in definition_query)

    def test_get_definition_of_library(self):
        definition_query = list(sdn.get_definitions(self.netlist.libraries[1]))
        self.assertTrue(len(definition_query) == 3 and self.leaf_inst.reference not in definition_query)

    def test_get_definition_of_library_outside(self):
        definition_query = list(sdn.get_definitions(self.netlist.libraries[0], selection="OUTSIDE"))
        self.assertTrue(len(definition_query) == 1 and self.leaf_inst.reference not in definition_query)

    def test_get_definition_of_library_outside_recursive(self):
        definition_query = list(sdn.get_definitions(self.netlist.libraries[0], selection="OUTSIDE", recursive=True))
        self.assertTrue(len(definition_query) == 3 and self.leaf_inst.reference not in definition_query)

    def test_get_definition_of_library_recursive(self):
        definition_query = list(sdn.get_definitions(self.netlist.libraries[1], recursive=True))
        self.assertTrue(len(definition_query) == 4 and self.leaf_inst.reference in definition_query)

    def test_get_definition_of_library_recursive_absolute_pattern(self):
        definition_query = list(sdn.get_definitions(self.netlist.libraries[1], "leaf", recursive=True))
        self.assertTrue(len(definition_query) == 1 and self.leaf_inst.reference is definition_query[0])

    def test_get_definition_of_definition_inside(self):
        definition_query = list(sdn.get_definitions(self.netlist.libraries[1].definitions[0],
                                                    selection="INSIDE"))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.netlist.libraries[0].definitions[0])

    def test_get_definition_of_definition_outside(self):
        definition_query = list(sdn.get_definitions(self.netlist.libraries[0].definitions[0],
                                               selection="OUTSIDE"))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.netlist.libraries[1].definitions[0])

    def test_get_definition_of_definition_inside_recursive(self):
        definition_query = list(sdn.get_definitions(self.netlist.top_instance.reference,
                                               selection="INSIDE", recursive=True))
        self.assertTrue(len(definition_query) == 3 and self.netlist.top_instance.reference not in definition_query)

    def test_get_definition_from_outer_pins(self):
        definition_query = list(sdn.get_definitions(self.leaf_inst.pins))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.leaf_inst.reference)

    def test_get_definition_from_inner_pins(self):
        definition_query = list(sdn.get_definitions(self.leaf_inst.reference.ports[0].pins[0]))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.leaf_inst.reference)

    def test_get_definition_from_wire(self):
        definition_query = list(sdn.get_definitions(self.netlist.top_instance.reference.cables[0].wires[0]))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.netlist.top_instance.reference)

    def test_get_definition_from_href(self):
        href = next(sdn.get_hinstances(self.netlist.top_instance))
        definition_query = list(sdn.get_definitions(href))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.netlist.top_instance.reference)

    def test_unique_query_return(self):
        definition_query = list(sdn.get_definitions([self.netlist.libraries[0], self.leaf_inst], "leaf"))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.leaf_inst.reference)

    def test_absolute_pattern_from_relative_query(self):
        definition_query = list(sdn.get_definitions(self.leaf_inst, "leaf"))
        self.assertTrue(len(definition_query) == 1 and definition_query[0] == self.leaf_inst.reference)



import unittest
import spydrnet as sdn


class TestGetHCables(unittest.TestCase):
    netlist = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.netlist = sdn.load_example_netlist_by_name('b13')

    def test_wild_card_search_on_netlist(self):
        hrefs = list(sdn.get_hcables(self.netlist))
        self.assertTrue(len(hrefs) == 76 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))

    def test_absolute_search(self):
        href = next(self.netlist.get_hcables('<const0>'), None)
        self.assertIsNotNone(href)
        self.assertIsInstance(href.item, sdn.Cable)

    def test_regex_search(self):
        hrefs = list(sdn.get_hcables(
            self.netlist, '.*FSM_onehot.*', is_re=True))
        self.assertTrue(len(hrefs) == 8 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))

    def test_parameters(self):
        self.assertRaises(TypeError, sdn.get_hcables, self.netlist,
                          r'.*FSM_onehot.*', patterns=r".*FSM_onehot.*")
        self.assertRaises(TypeError, sdn.get_hcables,
                          self.netlist, parameter_does_not_exit=True)
        self.assertRaises(TypeError, sdn.get_hcables, object())

    def test_get_hcables_of_pin(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        port = definition.ports[0]
        pin = port.pins[0]
        hrefs = list(sdn.get_hcables(pin))
        self.assertTrue(len(hrefs) == 1)
        self.assertIsInstance(hrefs[0].item, sdn.Cable)

    def test_get_hcables_of_wire(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        cable = definition.cables[0]
        wire = cable.wires[0]
        hrefs = list(sdn.get_hcables(wire))
        self.assertTrue(len(hrefs) == 1)
        self.assertIsInstance(hrefs[0].item, sdn.Cable)

    def test_get_hwire_of_invalid_reference(self):
        from spydrnet.util.hierarchical_reference import HRef
        invalid_href = HRef.from_parent_and_item(None, None)
        hrefs = list(sdn.get_hcables(invalid_href))
        self.assertTrue(len(hrefs) == 0)

    def test_get_hcables_from_hrefs_of_cable_and_wire(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        cable = definition.cables[0]
        wire = cable.wires[0]
        hrefs = list(sdn.get_hinstances(wire))
        href_top = hrefs[0]
        from spydrnet.util.hierarchical_reference import HRef
        cable_href = HRef.from_parent_and_item(href_top, cable)
        self.assertIsInstance(cable_href.item, sdn.Cable)
        wire_href = HRef.from_parent_and_item(cable_href, wire)
        href_result = next(sdn.get_hcables(cable_href), None)
        self.assertTrue(href_result is cable_href)
        href_result = next(sdn.get_hcables(wire_href), None)
        self.assertTrue(href_result is cable_href)

    def test_get_hcables_from_hrefs_of_port_and_pin(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        port = definition.ports[0]
        pin = port.pins[0]
        hrefs = list(sdn.get_hcables(pin))
        href = hrefs[0]
        self.assertIsInstance(href.item, sdn.Cable)
        from spydrnet.util.hierarchical_reference import HRef
        port_href = HRef.from_parent_and_item(href.parent, port)
        href_result = next(sdn.get_hcables(port_href), None)
        self.assertTrue(href_result is href)
        pin_href = HRef.from_parent_and_item(port_href, pin)
        href_result = next(sdn.get_hcables(pin_href), None)
        self.assertTrue(href_result is href)

    def test_from_href_of_instance(self):
        href = next(sdn.get_hinstances(self.netlist.top_instance))
        hrefs = list(sdn.get_hcables(href))
        self.assertTrue(len(hrefs) == 76 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))

    def test_from_instance(self):
        hrefs = list(sdn.get_hcables(self.netlist.top_instance))
        self.assertTrue(len(hrefs) == 76 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))

    def test_from_library(self):
        hrefs = list(sdn.get_hcables(
            self.netlist.libraries[0], selection=sdn.OUTSIDE))
        self.assertTrue(len(hrefs) == 76 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))

    def test_from_wire_and_cable(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        cable = definition.cables[0]
        wire = cable.wires[0]
        hrefs = list(sdn.get_hcables(wire))
        self.assertTrue(len(hrefs) == 1 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))
        hrefs = list(sdn.get_hcables(cable))
        self.assertTrue(len(hrefs) == 1 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))

    def test_from_outerpin(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        instance = definition.children[0]
        outerpin = next(iter(instance.pins))
        hrefs = list(sdn.get_hcables(outerpin, selection="OUTSIDE"))
        self.assertTrue(len(hrefs) == 1 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))

    def test_bad_selection_type(self):
        self.assertRaises(TypeError, self.netlist.get_hcables,
                          selection="NOT_AN_OPTION")
        self.assertRaises(TypeError, self.netlist.get_hcables, selection=None)

    def test_of_bad_instance(self):
        hrefs = list(sdn.get_hcables(sdn.Instance()))
        self.assertTrue(len(hrefs) == 0)

    def test_through_hierarchy(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        instance = sdn.Instance()
        instance.reference = definition
        netlist.top_instance = instance

        middle_inst = definition.create_child()
        middle_def = library.create_definition()
        middle_cable = middle_def.create_cable()
        middle_wire = middle_cable.create_wire()
        middle_inst.reference = middle_def

        leaf_inst = middle_def.create_child()
        leaf_def = library.create_definition()
        leaf_inst.reference = leaf_def

        top_inst_href = next(sdn.get_hinstances(netlist.top_instance))
        hrefs = list(sdn.get_hcables(top_inst_href, recursive=True))
        self.assertTrue(len(hrefs) == 1 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))

        hrefs = list(sdn.get_hcables(middle_inst))
        self.assertTrue(len(hrefs) == 1 and all(
            isinstance(x.item, sdn.Cable) for x in hrefs))

    def test_through_hierarchy_again(self):
        netlist = sdn.Netlist()

        library = netlist.create_library()
        library.name = 'work'

        leaf_def = library.create_definition()
        leaf_def.name = 'leaf'
        leaf_port = leaf_def.create_port()
        leaf_port.name = 'I'
        leaf_port.create_pins(1)

        bottom_def = library.create_definition()
        bottom_def.name = 'bottom'
        bottom_port = bottom_def.create_port()
        bottom_port.name = 'I'
        bottom_port.create_pins(1)
        leaf_inst = bottom_def.create_child()
        leaf_inst.reference = leaf_def
        bottom_cable = bottom_def.create_cable()
        bottom_cable.name = 'bottom_cable'
        bottom_wire = bottom_cable.create_wire()
        bottom_wire.connect_pin(bottom_port.pins[0])
        bottom_wire.connect_pin(leaf_inst.pins[leaf_port.pins[0]])

        bottom_floating_wire = bottom_cable.create_wire()

        middle_def = library.create_definition()
        middle_def.name = 'middle'
        middle_port = middle_def.create_port()
        middle_port.name = "I"
        middle_port.create_pin()
        bottom_inst = middle_def.create_child()
        bottom_inst.name = 'bottom'
        bottom_inst.reference = bottom_def
        middle_cable = middle_def.create_cable()
        middle_cable.name = "middle_cable"
        middle_wire = middle_cable.create_wire()
        middle_wire.connect_pin(middle_port.pins[0])
        middle_wire.connect_pin(bottom_inst.pins[bottom_port.pins[0]])

        middle_floating_wire = middle_cable.create_wire()

        top_def = library.create_definition()
        top_def.name = 'top'
        top_port = top_def.create_port()
        top_port.name = "I"
        top_port.create_pin()
        middle_inst = top_def.create_child()
        middle_inst.name = 'middle'
        middle_inst.reference = middle_def
        top_cable = top_def.create_cable()
        top_cable.name = "top_cable"
        top_wire = top_cable.create_wire()
        top_wire.connect_pin(top_port.pins[0])
        top_wire.connect_pin(middle_inst.pins[middle_port.pins[0]])

        top_floating_wire = top_cable.create_wire()

        top_instance = sdn.Instance()
        top_instance.name = 'top'
        top_instance.reference = top_def
        netlist.top_instance = top_instance

        href = next(sdn.get_hcables(top_floating_wire))
        hrefs = set(sdn.get_hcables(netlist.top_instance))
        self.assertTrue(href in hrefs)

        # look at wire_name
        href = next(sdn.get_hcables(middle_floating_wire))
        self.assertTrue('middle/middle_cable[1]', href.name)
        hrefs = set(sdn.get_hcables(netlist.top_instance, recursive=True))
        self.assertTrue(href in hrefs)

        hrefs = set(sdn.get_hcables(middle_cable, selection="OUTSIDE"))
        href_top_wire = next(sdn.get_hcables(top_cable.wires[0]))
        href_middle_wire = next(sdn.get_hcables(middle_cable.wires[0]))
        href_bottom_wire = next(sdn.get_hcables(bottom_cable.wires[0]))
        self.assertTrue(
            href_top_wire in hrefs and href_middle_wire not in hrefs and href_bottom_wire in hrefs)

        hrefs = set(sdn.get_hcables(middle_cable, selection="ALL"))
        href_middle_floating_wire = next(sdn.get_hcables(middle_floating_wire))
        self.assertTrue(href_top_wire in hrefs and href_middle_wire in hrefs and href_bottom_wire in hrefs and
                        href_middle_floating_wire in hrefs)

        all_wires = set(netlist.get_hcables())
        self.assertTrue(len(all_wires) == 1)

        all_wires = set(netlist.get_hcables(recursive=True))
        self.assertTrue(len(all_wires) == 3)



from spydrnet.ir import InnerPin, OuterPin, Wire, Netlist, Library, Definition, Cable, Element,\
    Instance, Port
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.selection import Selection
from spydrnet.global_state.global_service import lookup
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern


def get_cables(obj, *args, **kwargs):
    """
    get_cables(obj, ...)

    Get cables *within* an object.

    Parameters
    ----------    
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_cables(definition, ...)` would
        return all of the cables associated with the provided definition that match the additional criteria.
    patterns : str, Iterable - optional, positional or named, default: wildcard
        The search patterns. Patterns can be a single string or an Iterable collection of strings. Patterns can be
        absolute or they can contain wildcards or regular expressions. If `patterns` is not provided, then it defaults
        to a wildcard. Patterns are queried against the object property value stored under a specified key. Fast lookups
        are only attempted on absolute patterns that are not regular expressions and contain no wildcards.
    key : str, optional, default: ".NAME"
        This is the key that controls which value is being searched.
    is_case : bool - optional, named, default: True
        Specify if patterns should be treated as case sensitive. Only applies to patterns. Does not alter fast lookup
        behavior (if namespace policy uses case insensitive indexing, this parameter will not prevent a fast lookup
        from returning a matching object even if the case is not an exact match).
    is_re: bool - optional, named, default: False
        Specify if patterns are regular expressions. If `False`, a pattern can still contain `*` and `?` wildcards. A
        `*` matches zero or more characters. A `?` matches upto a single character.
    selection : Selection.{INSIDE, OUTSIDE, BOTH, ALL}, default: INSIDE
        This parameter determines the wires that are returned based on the instance associated with the object that is
        being searched.
    recursive : bool - optional, default: False
        Specify if search should be recursive or not meaning that sub hierarchical instances within an instance are
        included or not.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    cables : generator
        The cables associated with a particular object or collection of objects.
    
    """
    # Check argument list
    if len(args) == 1 and 'patterns' in kwargs:
        raise TypeError("get_cables() got multiple values for argument 'patterns'")
    if len(args) > 1 or any(x not in {'patterns', 'key', 'filter', 'is_case', 'is_re', 'selection', 'recursive'}
                            for x in kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    selection = kwargs.get('selection', Selection.INSIDE)
    if isinstance(selection, str):
        if selection in Selection.__members__:
            selection = Selection[selection]
    if isinstance(selection, Selection) is False:
        raise TypeError("selection must be '{}'".format("', '".join(Selection.__members__.keys())))

    filter_func = kwargs.get('filter', lambda x: True)
    is_case = kwargs.get('is_case', True)
    is_re = kwargs.get('is_re', False)
    patterns = args[0] if len(args) == 1 else kwargs.get('patterns', ".*" if is_re else "*")
    key = kwargs.get('key', ".NAME")
    recursive = kwargs.get('recursive', False)

    if isinstance(obj, (Element, HRef)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (Element, HRef)) for x in object_collection) is False:
        raise TypeError("get_cables() only supports netlists, libraries, and definitions, or a collection of these as "
                        "the object searched")

    if isinstance(patterns, str):
        patterns = (patterns,)

    return _get_cables(object_collection, patterns, key, is_case, is_re, selection, recursive, filter_func)


def _get_cables(object_collection, patterns, key, is_case, is_re, selection, recursive, filter_func):
    for result in filter(filter_func, _get_cables_raw(object_collection, patterns, key, is_case, is_re, selection,
                                                      recursive)):
        yield result


def _get_cables_raw(object_collection, patterns, key, is_case, is_re, selection, recursive):
    found = set()
    other_cables = set()
    searched_wires = set()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Definition):
            if selection in {Selection.INSIDE, Selection.ALL}:
                for pattern in patterns:
                    pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
                    if pattern_is_absolute:
                        result = lookup(obj, Cable, key, pattern)
                        if result is not None and result not in found:
                            found.add(result)
                            yield result
                    else:
                        for cable in obj.cables:
                            value = cable[key] if key in cable else ''
                            if cable not in found and _value_matches_pattern(value, pattern, is_case, is_re):
                                found.add(cable)
                                yield cable
                if recursive or selection == Selection.ALL:
                    object_collection += obj.children
            if selection in {Selection.OUTSIDE, Selection.BOTH, Selection.ALL}:
                for port in obj.ports:
                    for pin in port.pins:
                        object_collection.append(pin)
        elif isinstance(obj, Library):
            object_collection += obj.definitions
        elif isinstance(obj, Netlist):
            for library in obj.libraries:
                object_collection += library.definitions
        elif isinstance(obj, Instance):
            if selection in {Selection.INSIDE, Selection.ALL}:
                reference = obj.reference
                if reference:
                    for cable in reference.cables:
                        other_cables.add(cable)
                    if recursive or selection == Selection.ALL:
                        object_collection += reference.children
            if selection in {Selection.BOTH, Selection.OUTSIDE, Selection.ALL}:
                object_collection += obj.pins
        elif isinstance(obj, InnerPin):
            if selection in {Selection.INSIDE, Selection.BOTH, Selection.ALL}:
                wire = obj.wire
                if wire and wire not in searched_wires:
                    searched_wires.add(wire)
                    cable = wire.cable
                    if cable:
                        other_cables.add(cable)
                    if selection == Selection.ALL:
                        object_collection += wire.pins
            if selection in {Selection.OUTSIDE, Selection.BOTH, Selection.ALL}:
                port = obj.port
                if port:
                    definition = port.definition
                    if definition:
                        for instance in definition.references:
                            outer_pin = instance.pins[obj]
                            outer_wire = outer_pin.wire
                            if outer_wire and outer_wire not in searched_wires:
                                searched_wires.add(outer_wire)
                                cable = outer_wire.cable
                                if cable:
                                    other_cables.add(cable)
                                if selection == Selection.ALL:
                                    object_collection += outer_wire.pins
        elif isinstance(obj, OuterPin):
            if selection in {Selection.INSIDE, Selection.BOTH, Selection.ALL}:
                inner_pin = obj.inner_pin
                if inner_pin:
                    inner_wire = inner_pin.wire
                    if inner_wire and inner_wire not in searched_wires:
                        searched_wires.add(inner_wire)
                        cable = inner_wire.cable
                        if cable:
                            other_cables.add(cable)
                        if selection == Selection.ALL:
                            object_collection += inner_wire.pins
            if selection in {Selection.OUTSIDE, Selection.BOTH, Selection.ALL}:
                wire = obj.wire
                if wire and wire not in searched_wires:
                    searched_wires.add(wire)
                    cable = wire.cable
                    if cable:
                        other_cables.add(cable)
                    if selection == Selection.ALL:
                        object_collection += wire.pins
        elif isinstance(obj, Port):
            object_collection += obj.pins
        elif isinstance(obj, Wire):
            if selection == Selection.INSIDE:
                cable = obj.cable
                if cable:
                    other_cables.add(cable)
            elif selection == Selection.OUTSIDE:
                for pin in obj.pins:
                    if isinstance(pin, OuterPin):
                        inner_pin = pin.inner_pin
                        inner_wire = inner_pin.wire
                        if inner_wire:
                            cable = inner_wire.cable
                            if cable:
                                other_cables.add(cable)
                    else:
                        object_collection.append(pin)
            else:
                object_collection += obj.pins
        elif isinstance(obj, Cable):
            if selection == Selection.INSIDE:
                other_cables.add(obj)
            else:
                object_collection += obj.wires
        elif isinstance(obj, HRef):
            if obj.is_valid:
                object_collection.append(obj.item)

    if other_cables:
        namemap = dict()
        for other_cable in other_cables:
            if other_cable not in found:
                found.add(other_cable)
                name = other_cable[key] if key in other_cable else ''
                if name not in namemap:
                    namemap[name] = list()
                namemap[name].append(other_cable)
        for pattern in patterns:
            pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
            if pattern_is_absolute:
                if pattern in namemap:
                    result = namemap[pattern]
                    del namemap[pattern]
                    for port in result:
                        yield port
            else:
                names_to_remove = list()
                for name in namemap:
                    if _value_matches_pattern(name, pattern, is_case, is_re):
                        result = namemap[name]
                        names_to_remove.append(name)
                        for port in result:
                            yield port
                for name in names_to_remove:
                    del namemap[name]



# Copyright 2021 please see the license
# Author Dallin Skouson

import unittest
import spydrnet as sdn
from spydrnet.parsers.verilog.parser import VerilogParser
import spydrnet.parsers.verilog.verilog_tokens as vt
from spydrnet import parsers
import os

class TestVerilogParser(unittest.TestCase):

    class TestTokenizer:
        def __init__(self, tokenlist):
            self.tokens = tokenlist
            self.index = 0
            self.highest = len(tokenlist)

        @property
        def line_number(self):
            return self.index

        def has_next(self):
            return self.index < self.highest

        def next(self):
            token = self.tokens[self.index]
            self.index += 1
            return token
        
        def peek(self):
            token = self.tokens[self.index]
            return token

    ###################################################
    ##Module Headers
    ###################################################

    def test_module_header_parameter_parsing(self):
        expected = dict()
        expected["INIT"] = "1'h1"
        expected["[1:0] INIT0"] = "2'h0"
        expected["[0] INIT1"] = "1'b0"
        tokens = ["#", "(", "parameter", "INIT", "=", "1'h1", ",",\
            "parameter", "[", "1", ":", "0", "]", "INIT0", "=", "2'h0", ",",\
            "parameter", "[", "0", "]", "INIT1", "=", "1'b0", ")"]

        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        parser.current_definition = sdn.Definition()

        parser.parse_module_header_parameters()

        assert "VERILOG.Parameters" in parser.current_definition, "expected parameters in the definition"
        parameters = parser.current_definition["VERILOG.Parameters"]
        for k, v in expected.items():
            assert k in parameters, "expected to see " + k + " in the definition parameters"
            assert parameters[k] == v, "expected value of k to be " + v + " but got instead " + parameters[k]
        
        for k, v in parameters.items():
            assert k in expected, "unexpected value " + k + " in the definition parameters"
        
    def test_module_header_port_name_only(self):
        #expected = [(direction, left, right, name)]
        expected = ["PORT0", "PORT1", "PORT2", "PORT3", "PORT4"]
        tokens = ["("]
        first_run = True
        for p in expected:
            if first_run:
                tokens.append(p)
                first_run = False
            else:
                tokens.append(",")
                tokens.append(p)
        tokens.append(")")

        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.current_definition = sdn.Definition()
        parser.tokenizer = tokenizer

        parser.parse_module_header_ports()

        assert len(parser.current_definition.ports) == len(expected), \
            "port count mismatch definition: " + str(len(parser.current_definition.ports)) + " expected: " + str(len(expected))

        for i in range(len(expected)):
            assert expected[i] == parser.current_definition.ports[i].name

    def test_module_header_port_name_and_direction(self):
        #expected = [(direction, left, right, name)]
        expected = [("PORT0", "output"), ("PORT1", "input"), ("PORT2", "inout"), ("PORT3", "output"), ("PORT4", "input")]
        tokens = ["("]
        first_run = True
        for p in expected:
            if first_run:
                tokens.append(p[1])
                tokens.append(p[0])
                first_run = False
            else:
                tokens.append(",")
                tokens.append(p[1])
                tokens.append(p[0])
        tokens.append(")")

        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.current_definition = sdn.Definition()
        parser.tokenizer = tokenizer

        parser.parse_module_header_ports()

        assert len(parser.current_definition.ports) == len(expected), \
            "port count mismatch definition: " + str(len(parser.current_definition.ports)) + " expected: " + str(len(expected))

        for i in range(len(expected)):
            assert expected[i][0] == parser.current_definition.ports[i].name,\
                "ports names don't match. Definition: " + parser.current_definition.ports[i].name + " expected " + expected[i][0]
            expected_direction = vt.string_to_port_direction(expected[i][1])
            definition_direction = parser.current_definition.ports[i].direction
            assert expected_direction == definition_direction, \
                "directions do not match up expected " + str(expected_direction) + " but got " + definition_direction

    def test_module_heaader_port_name_and_index(self):
        #expected = [(direction, left, right, name)]
        expected = [("PORT0", "5" , "0"), ("PORT1", "8","-2"), ("PORT2", "0","0"), ("PORT3", "16","8"), ("PORT4", "3","0")]
        tokens = ["("]
        first_run = True
        for p in expected:
            if first_run:
                first_run = False
            else:
                tokens.append(",")
            tokens.append("[")
            tokens.append(p[1])
            tokens.append(":")
            tokens.append(p[2])
            tokens.append("]")
            tokens.append(p[0])
        tokens.append(")")

        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.current_definition = sdn.Definition()
        parser.tokenizer = tokenizer

        parser.parse_module_header_ports()

        assert len(parser.current_definition.ports) == len(expected), \
            "port count mismatch definition: " + str(len(parser.current_definition.ports)) + " expected: " + str(len(expected))

        for i in range(len(expected)):
            assert expected[i][0] == parser.current_definition.ports[i].name,\
                "ports names don't match. Definition: " + parser.current_definition.ports[i].name + " expected " + expected[i][0]
            expected_lower = min(int(expected[i][2]), int(expected[i][1]))
            definition_lower = parser.current_definition.ports[i].lower_index

            expected_width = max(int(expected[i][2]), int(expected[i][1])) - expected_lower + 1
            definition_width = len(parser.current_definition.ports[i].pins)
            assert expected_lower == definition_lower, \
                "lower indicies do not match up expected " + str(expected_lower) + " but got " + str(definition_lower)
            
            assert expected_width == definition_width, \
                "widths do not match up expected " + str(expected_width) + " but got " + str(definition_width)


    def test_module_header_all_present(self):
        expected = [("PORT0", "5" , "0", "input"), ("PORT1", "8","-2", "output"), ("PORT2", "0","0", "inout"),\
            ("PORT3", "16","8", "input"), ("PORT4", "3","0", "output")]
        tokens = ["("]
        first_run = True
        for p in expected:
            if first_run:
                first_run = False
            else:
                tokens.append(",")
            tokens.append(p[3])
            tokens.append("[")
            tokens.append(p[1])
            tokens.append(":")
            tokens.append(p[2])
            tokens.append("]")
            tokens.append(p[0])
        tokens.append(")")

        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.current_definition = sdn.Definition()
        parser.tokenizer = tokenizer

        parser.parse_module_header_ports()

        assert len(parser.current_definition.ports) == len(expected), \
            "port count mismatch definition: " + str(len(parser.current_definition.ports)) + " expected: " + str(len(expected))

        for i in range(len(expected)):
            assert expected[i][0] == parser.current_definition.ports[i].name,\
                "ports names don't match. Definition: " + parser.current_definition.ports[i].name + " expected " + expected[i][0]
            expected_lower = min(int(expected[i][2]), int(expected[i][1]))
            definition_lower = parser.current_definition.ports[i].lower_index

            expected_width = max(int(expected[i][2]), int(expected[i][1])) - expected_lower + 1
            definition_width = len(parser.current_definition.ports[i].pins)
            assert expected_lower == definition_lower, \
                "lower indicies do not match up expected " + str(expected_lower) + " but got " + str(definition_lower)
            
            assert expected_width == definition_width, \
                "widths do not match up expected " + str(expected_width) + " but got " + str(definition_width)

            expected_direction = vt.string_to_port_direction(expected[i][3])
            definition_direction = parser.current_definition.ports[i].direction
            assert expected_direction == definition_direction, \
                "directions do not match up expected " + str(expected_direction) + " but got " + definition_direction

    def test_port_aliasing(self):
        parser = VerilogParser()
        token_list = [".", "portName", "(", "{", "cable1", ",", "cable2", ",", "cable3", "}", ")"]
        parser.tokenizer = self.TestTokenizer(token_list)

        d = sdn.Definition()
        parser.current_definition = d

        parser.parse_module_header_port_alias()
        assert len(d.ports) == 1, "definition has the wrong number of ports, should have 1 has " + str(len(d.ports))
        assert d.ports[0].name == "portName", "port has the wrong name it should be portName but is " + d.ports[0].name
        assert len(d.ports[0].pins) == 3
        assert len(d.cables) == 3
        names = [d.cables[0].name, d.cables[1].name, d.cables[2].name]
        assert "cable1" in names and "cable2" in names and "cable3" in names

    def test_port_aliasing_single(self):
        parser = VerilogParser()
        token_list = [".", "portName", "(", "cable1", ")"]
        parser.tokenizer = self.TestTokenizer(token_list)

        d = sdn.Definition()
        parser.current_definition = d

        parser.parse_module_header_port_alias()

        assert len(d.ports) == 1
        assert d.ports[0].name == "portName"
        assert len(d.cables) == 1
        assert len(d.ports[0].pins) == 1
        assert len(d.cables[0].wires) == 1
        assert d.cables[0].name == "cable1"

    def test_non_zero_lower_port_index(self):
        '''this test based on an issue found while parsing riscv_multi_core.v in the support files
        to see the problem lines that caused this test see line 89 and line 41'''

        parser = VerilogParser()
        token_list = ["module", "alu", "(", "instruction", ")", ";", "output", "[", "27", ":", "26", "]", "instruction", ";", "endmodule"]
        tokenizer = self.TestTokenizer(token_list)
        parser.tokenizer = tokenizer
        parser.current_library = sdn.Library(name = "TestLibrary")
        parser.netlist = sdn.Netlist()
        parser.netlist.add_library(parser.current_library)
        
        parser.parse_module()

        assert len(parser.current_definition.ports) == 1
        assert len(parser.current_definition.cables) == 1
        c = parser.current_definition.cables[0]
        p = parser.current_definition.ports[0]
        assert c.name == "instruction"
        assert len(c.wires) == 2
        assert p.name == "instruction"
        assert len(p.pins) == 2
        assert c.lower_index == 26
        assert p.lower_index == 26

        for w in c.wires:
            assert w.pins[0] in p.pins

    def test_port_cable_resize_post_creation(self):
        '''we need a test that creates a port and then resizes it on definition
        
        module something(port);
        input [3:0] port;
        endmodule'''
        parser = VerilogParser()
        token_list = ["module", "alu", "(", "instruction", ")", ";", "output", "[", "3", ":", "0", "]", "instruction", ";", "endmodule"]
        tokenizer = self.TestTokenizer(token_list)
        parser.tokenizer = tokenizer
        parser.current_library = sdn.Library(name = "TestLibrary")
        parser.netlist = sdn.Netlist()
        parser.netlist.add_library(parser.current_library)
        
        parser.parse_module()

        assert len(parser.current_definition.ports) == 1
        assert len(parser.current_definition.cables) == 1
        c = parser.current_definition.cables[0]
        p = parser.current_definition.ports[0]
        assert c.name == "instruction"
        assert len(c.wires) == 4
        assert p.name == "instruction"
        assert len(p.pins) == 4
        assert c.lower_index == 0
        assert p.lower_index == 0

        for w in c.wires:
            assert w.pins[0] in p.pins

    def test_port_resize_on_aliased_port(self):
        '''example of this found in sha3_core.v on line 63176 or 63177 something to do with \\byte_num[0] '''
        tokens = ["module", "keccak", "(",\
            '.','byte_num','(','{','\\byte_num[2] ',',','\\byte_num[1] ', ',','\\byte_num[0] ','}',')',\
            ")", ";", "input", '\\byte_num[0] ', ';', 'input', '\\byte_num[1] ', ';', 'input', '\\byte_num[2] ',';', "endmodule"]
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        parser.current_library = sdn.Library(name = "TestLibrary")
        parser.netlist = sdn.Netlist()
        parser.netlist.add_library(parser.current_library)
        
        parser.parse_module()

        assert len(parser.current_definition.ports) == 1
        assert len(parser.current_definition.cables) == 3
        names = []
        for c in parser.current_definition.cables:
            names.append(c.name)
            assert len(c.wires) == 1
        assert '\\byte_num[2] ' in names
        assert '\\byte_num[1] ' in names
        assert '\\byte_num[0] ' in names
        assert parser.current_definition.ports[0].name == 'byte_num'



    def test_parse_empty_module_header(self):
        '''example in bram.v in the support files on line 18'''
        tokens = ["(", ")", ";"]
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        parser.current_definition = sdn.Definition

        parser.parse_module_header() #intent is to just make sure this does not crash
        #todo add some aserts to check to make sure the number of cables and ports is still 0


    ###################################################
    ##Array Slicing
    ###################################################

    def test_array_slicing_good_single(self):
        left_expected = 2334
        right_expected = None
        tokens = ["[", str(left_expected), "]"]
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        left, right = parser.parse_brackets()
        assert(left == left_expected)
        assert(right == right_expected)

    def test_array_slicing_good_both(self):
        left_expected = 2334
        right_expected = 231
        tokens = ["[", str(left_expected), ":", str(right_expected), "]"]
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        left, right = parser.parse_brackets()
        assert(left == left_expected)
        assert(right == right_expected)

    @unittest.expectedFailure
    def test_array_slicing_bad_colon(self):
        left_expected = 2334
        right_expected = 231
        tokens = ["[", str(left_expected), ",", str(right_expected), "]"]
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        left, right = parser.parse_brackets()
        assert(left == left_expected)
        assert(right == right_expected)
    
    @unittest.expectedFailure
    def test_array_slicing_bad_ending(self):
        left_expected = 2334
        right_expected = 231
        tokens = ["[", str(left_expected), ":", str(right_expected), "something_else"]
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        left, right = parser.parse_brackets()
        assert(left == left_expected)
        assert(right == right_expected)

    @unittest.expectedFailure
    def test_array_slicing_bad_non_numbers(self):
        left_expected = "not a number"
        right_expected = None
        tokens = ["[", str(left_expected), ":", str(right_expected), "something_else"]
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        left, right = parser.parse_brackets()
        assert(left == left_expected)
        assert(right == right_expected)
        

    ###################################################
    ##Cable Creation and modification
    ###################################################

    def init_cable_creation(self):
        parser = VerilogParser()
        
        my_def = sdn.Definition()
        my_def.name = "testing_definition"
        c1 = my_def.create_cable()
        c1.name = "preexisting1" #[15:8]
        c1.create_wires(8)
        c1.lower_index = 8
        c2 = my_def.create_cable()
        c2.name = "preexisting2" #[15:0]
        c2.create_wires(16)
        c3 = my_def.create_cable()
        c3.name = "preexisting3" #[19:16]
        c3.is_downto = False
        c3.lower_index = 16
        c3.create_wires(4)
        
        parser.current_definition = my_def

        return parser, c1, c2, c3

    def add_to_cable_helper(self, tests, parser):
        #tests is a list of tuples that contain
        #name,left,right,type,cable,expected_lower,expected_width, expected_downto

        for entry in tests:
            name, left_index, right_index, var_type, cable, expected_lower, expected_width = entry
            expected_downto = cable.is_downto #the cable should not update it's downto once that has been decided.
            c = parser.create_or_update_cable(name, left_index, right_index, var_type)
            assert c is cable, "the cable is not the expected cable, " + name
            assert len(c.wires) == expected_width, "the cable is not the proper width, " + name
            assert c.lower_index == expected_lower, "the cable does not have the proper lower index, " + name
            assert c.is_downto == expected_downto, "the cable changed downto, " + name


    def create_cable_helper(self, tests, parser):
        #tests is a list of tuples that contain
        #name,left,right,type,cable,expected_lower,expected_width, expected_downto
        start_size = len(parser.current_definition.cables)
        count = 0
        for entry in tests:
            count += 1
            name, left_index, right_index, var_type, cable, expected_lower, expected_width, expected_downto = entry
            c = parser.create_or_update_cable(name, left_index, right_index, var_type)
            assert len(c.wires) == expected_width, "the cable is not the proper width, " + name
            assert c.lower_index == expected_lower, "the cable does not have the proper lower index, " + name
            assert c.is_downto == expected_downto
        assert len(parser.current_definition.cables) == count + start_size, "the wrong number of cables were added"


    def test_add_to_front_of_cable(self):
        parser, c1, c2, c3 = self.init_cable_creation()
        
        tests =\
            [("preexisting1", 15, 0, None, c1, 0, 16),\
            ("preexisting2", 0, -16, None, c2, -16, 32),\
            ("preexisting3", 17, 0, None, c3, 0, 20)]
        self.add_to_cable_helper(tests, parser)


    def test_add_to_back_of_cable(self):
        parser, c1, c2, c3 = self.init_cable_creation()
        
        tests =\
            [("preexisting1", 31, 16, None, c1, 8, 24),\
            ("preexisting2", 31, 8, None, c2, 0, 32),\
            ("preexisting3", 63, 0, None, c3, 0, 64)]
        self.add_to_cable_helper(tests, parser)


    def test_create_cable(self):
        parser, c1, c2, c3 = self.init_cable_creation()
        
        tests =\
            [("new_cable1", 31, 16, None, None, 16, 16, True),\
            ("new_cable2", 0, -7, None, None, -7, 8, True),\
            ("new_cable3", 63, 0, None, None, 0, 64, True)]
        self.create_cable_helper(tests, parser)


    def test_create_cable_downto(self):
        parser, c1, c2, c3 = self.init_cable_creation()
        
        tests =\
            [("new_cable1", 7, 0, None, None, 0, 8, True),\
            ("new_cable2", 0, 7, None, None, 0, 8, False)]
        self.create_cable_helper(tests, parser)

    def test_change_cable_type(self):
        pass #TODO

    def test_dont_change_cable(self):
        parser, c1, c2, c3 = self.init_cable_creation()
        
        tests =\
            [("preexisting1", 15, 8, None, c1, 8, 8),\
            ("preexisting2", 15, 0, None, c2, 0, 16),\
            ("preexisting3", 19, 16, None, c3, 16, 4)]
        self.add_to_cable_helper(tests, parser)

    def test_single_index_cable(self):
        parser, c1, c2, c3 = self.init_cable_creation()
        tests_create =\
            [("new_cable1", 31, None, None, None, 31, 1, True),\
            ("new_cable2", 0, None, None, None, 0, 1, True),\
            ("new_cable3", 63, None, None, None, 63, 1, True)]
        self.create_cable_helper(tests_create, parser)

        parser, c1, c2, c3 = self.init_cable_creation()
        tests_modify =\
            [("preexisting1", 0, None, None, c1, 0, 16),\
            ("preexisting2", 8, None, None, c2, 0, 16),\
            ("preexisting3", 0, None, None, c3, 0, 20)]
        self.add_to_cable_helper(tests_modify, parser)


    def test_change_cable_downto(self):
        #i don't think the cable downto-ness should ever change. this is tested in the add to cable helper
        parser, c1, c2, c3 = self.init_cable_creation()
        
        tests =\
            [("preexisting1", 15, 8, None, c1, 8, 8),\
            ("preexisting1", 8, 15, None, c1, 8, 8)]
        self.add_to_cable_helper(tests, parser)

    def test_cable_prepend_wires(self):
        parser, c1, c2, c3 = self.init_cable_creation()
        c1_width = len(c1.wires)
        c1_lower = c1.lower_index
        c1_wires = set()
        for wire in c1.wires:
            c1_wires.add(wire)
        c2_width = len(c2.wires)
        c2_lower = c2.lower_index
        c2_wires = set()
        for wire in c2.wires:
            c2_wires.add(wire)
        c3_width = len(c3.wires)
        c3_lower = c3.lower_index
        c3_wires = set()
        for wire in c3.wires:
            c3_wires.add(wire)

        print(c1_wires)
        print()
        
        prepend_count = 8

        parser.prepend_wires(c1, prepend_count)
        print(c1_wires)
        print()
        print(c1.wires)
        print()
        assert len(c1.wires) == prepend_count + c1_width, "tried to add 8 wires to c1"
        assert c1.lower_index == c1_lower - prepend_count
        for wire in c1.wires[:prepend_count]:
            assert wire not in c1_wires, "a wire was in the wrong location for the prepend"
        for wire in c1.wires[prepend_count:]:
            assert wire in c1_wires, "a wire seems to have disappeared"

        parser.prepend_wires(c2, prepend_count)
        assert len(c2.wires) == prepend_count + c2_width, "tried to add 8 wires to c2"
        assert c2.lower_index == c2_lower - prepend_count
        for wire in c2.wires[:prepend_count]:
            assert wire not in c2_wires, "a wire was in the wrong location for the prepend"
        for wire in c2.wires[prepend_count:]:
            assert wire in c2_wires, "a wire seems to have disappeared"

        parser.prepend_wires(c3, prepend_count)
        assert len(c3.wires) == prepend_count + c3_width, "tried to add 8 wires to c3"
        assert c3.lower_index == c3_lower - prepend_count
        for wire in c3.wires[:prepend_count]:
            assert wire not in c3_wires, "a wire was in the wrong location for the prepend"
        for wire in c3.wires[prepend_count:]:
            assert wire in c3_wires, "a wire seems to have disappeared"

    def test_cable_postpend_wires(self):
        parser, c1, c2, c3 = self.init_cable_creation()
        c1_width = len(c1.wires)
        c1_lower = c1.lower_index
        c1_wires = set()
        for wire in c1.wires:
            c1_wires.add(wire)
        c2_width = len(c2.wires)
        c2_lower = c2.lower_index
        c2_wires = set()
        for wire in c2.wires:
            c2_wires.add(wire)
        c3_width = len(c3.wires)
        c3_lower = c3.lower_index
        c3_wires = set()
        for wire in c3.wires:
            c3_wires.add(wire)
        
        postpend_count = 8

        parser.postpend_wires(c1, postpend_count)
        assert len(c1.wires) == postpend_count + c1_width, "tried to add 8 wires to c1"
        assert c1.lower_index == c1_lower
        for wire in c1.wires[c1_width:]:
            assert wire not in c1_wires, "a wire was in the wrong location for the postpend"
        for wire in c1.wires[:c1_width]:
            assert wire in c1_wires, "a wire seems to have disappeared"

        parser.postpend_wires(c2, postpend_count)
        assert len(c2.wires) == postpend_count + c2_width, "tried to add 8 wires to c2"
        assert c2.lower_index == c2_lower
        for wire in c2.wires[c2_width:]:
            assert wire not in c2_wires, "a wire was in the wrong location for the postpend"
        for wire in c2.wires[:c2_width]:
            assert wire in c2_wires, "a wire seems to have disappeared"

        parser.postpend_wires(c3, postpend_count)
        assert len(c3.wires) == postpend_count + c3_width, "tried to add 8 wires to c3"
        assert c3.lower_index == c3_lower
        for wire in c3.wires[c3_width:]:
            assert wire not in c3_wires, "a wire was in the wrong location for the postpend"
        for wire in c3.wires[:c3_width]:
            assert wire in c3_wires, "a wire seems to have disappeared"

    def test_populate_new_cable(self):
        parser = VerilogParser()
        
        test_data = [("test_name", 2, 0, "wire"), ("\\escaped#$@_[213] ", 15,8,"wire"), ("some_name", 7,4,"reg")]

        for name, left, right, cable_type in test_data:
            cable = sdn.Cable()
            parser.populate_new_cable(cable,name,left,right,cable_type)
            assert cable.name == name
            assert cable.is_downto == (left >= right)
            assert cable["VERILOG.CableType"] == cable_type
            assert cable.lower_index == min(left,right)

    def test_parse_variable_instantiation(self):
        pass

    def test_parse_cable_concatenation(self):
        parser = VerilogParser()
        
        wire_names = ["wire1", "wire2", "wire3", "wire4"]
        token_list = ["{"]
        first = True
        for wn in wire_names:
            if first:
                first = False
            else:
                token_list.append(",")
            token_list.append(wn)
        token_list.append("}")

        tokenizer = self.TestTokenizer(token_list)
        parser.tokenizer = tokenizer
        parser.current_definition = sdn.Definition()

        wires = parser.parse_cable_concatenation()

        assert len(wires) == len(wire_names), "expected " + str(len(wire_names)) +  " wires to be created instead got " + str(len(wires))
        assert len(parser.current_definition.cables) == len(wire_names), "expected " + str(len(wire_names)) +  " cables to be created instead got " + str(len(parser.current_definition.cables))

        for i in range(len(wire_names)):
            assert parser.current_definition.cables[i].name == wire_names[i], " the wires created do not have matching names"
            assert wires[i].cable.name == wire_names[i], "the wires returned are not in order."
            
    ############################################################################
    ##Instance instantiation
    ############################################################################

    def test_parse_instantiation(self):
        pass

    def test_parse_empty_port_map(self):
        parser = VerilogParser()
        definition = sdn.Definition()
        parser.current_definition = definition
        parser.current_instance = definition.create_child()
        parser.current_instance.reference = parser.blackbox_holder.get_blackbox("definition_name")
        
        port_name = "port_name"
        
        tokens = [".", port_name, "(", ")"]

        tokenizer = self.TestTokenizer(tokens)

        parser.tokenizer = tokenizer

        parser.parse_port_map_single()

        assert len(parser.current_instance.reference.ports) == 1
        assert parser.current_instance.reference.ports[0].name == port_name
        assert len(parser.current_instance.reference.ports[0].pins) == 1

        assert tokenizer.has_next() == False

    def test_parse_parameter_map(self):
        parser = VerilogParser()
        tokens = ["#", "(", ".", "INIT", "(", "1'h123210", ")", ",", ".", "PARAM2", "(", "1'b0", ")", ")"]
        tokenizer = self.TestTokenizer(tokens)
        parser.tokenizer = tokenizer

        parser.current_definition = sdn.Definition()

        params = parser.parse_parameter_mapping()

        assert len(params.keys()) == 2
        assert params['INIT'] == "1'h123210"
        assert params["PARAM2"] == "1'b0"

    def test_parse_port_map(self):
        parser = VerilogParser()
        tokens = ["("]
        port_names = ["port1", "port2", "port3"]
        cable_names = ["cable1", "cable2", "cable3"]

        first = True
        for i in range(len(port_names)):
            if first:
                first = False
            else:
                tokens.append(",")
            tokens.append(".")
            tokens.append(port_names[i])
            tokens.append("(")
            tokens.append(cable_names[i])
            tokens.append(")")
        tokens.append(")")

        tokenizer = self.TestTokenizer(tokens)
        parser.tokenizer = tokenizer

        parser.current_definition = sdn.Definition()
        parser.current_instance = parser.current_definition.create_child()
        parser.current_instance.reference = parser.blackbox_holder.get_blackbox("definition1")

        parser.parse_port_mapping()

        assert len(parser.current_instance.reference.ports) == len(port_names)
        assert len(parser.current_instance.pins) == len(port_names)
        assert len(parser.current_definition.cables) == len(cable_names)
        for p in parser.current_instance.reference.ports:
            assert len(p.pins) == 1
        for c in parser.current_definition.cables:
            assert len(c.wires) ==1
        for i in range(len(port_names)):
            assert parser.current_definition.cables[i].name == cable_names[i]
            assert parser.current_instance.reference.ports[i].name == port_names[i]
            for p in parser.current_definition.cables[i].wires[0].pins:
                assert p in parser.current_instance.pins
            

    def test_parse_parameter_map_single(self):
        parser = VerilogParser()
        tokens = [".","INIT", "(", "1hABCD1230", ")"]
        tokenizer = self.TestTokenizer(tokens)
        parser.tokenizer = tokenizer

        k,v = parser.parse_parameter_map_single()

        assert k == "INIT"
        assert v == "1hABCD1230"
    

    def test_parse_parameter_map_single_multi_token_value(self):
        parser = VerilogParser()
        tokens = [".","INIT", "(", "200.000000", ")"]
        tokenizer = self.TestTokenizer(tokens)
        parser.tokenizer = tokenizer

        k,v = parser.parse_parameter_map_single()

        assert k == "INIT"
        assert v == "200.000000"
        assert parser.tokenizer.has_next() == False

    def test_parse_port_map_single(self):
        parser = VerilogParser()
        definition = sdn.Definition()
        parser.current_definition = definition
        parser.current_instance = definition.create_child()
        parser.current_instance.reference = parser.blackbox_holder.get_blackbox("definition_name")
        cable = definition.create_cable()
        cable_name = "cable_name"
        cable_width = 4
        port_name = "port_name"
        cable.name = cable_name
        cable.lower_index = 0
        cable.create_wires(cable_width)
        tokens = [".", port_name, "(", cable_name, "[", "3", ":", "0", "]", ")"]

        tokenizer = self.TestTokenizer(tokens)

        parser.tokenizer = tokenizer

        parser.parse_port_map_single()

        assert len(parser.current_instance.reference.ports) == 1
        assert parser.current_instance.reference.ports[0].name == port_name
        assert len(parser.current_instance.reference.ports[0].pins) == cable_width
        w_pins = []
        for w in cable.wires:
            assert w.pins[0] in parser.current_instance.pins
            w_pins.append(w.pins[0])
        for p in parser.current_instance.pins:
            assert p in w_pins

    ############################################################################
    ##Port creation and modification
    ############################################################################

    def init_port_creation(self):
        parser = VerilogParser()
        
        my_def = sdn.Definition()
        my_def.name = "testing_definition"
        c1 = my_def.create_port()
        c1.name = "preexisting1" #[15:8]
        c1.create_pins(8)
        c1.lower_index = 8
        c2 = my_def.create_port()
        c2.name = "preexisting2" #[15:0]
        c2.create_pins(16)
        c3 = my_def.create_port()
        c3.name = "preexisting3" #[19:16]
        c3.is_downto = False
        c3.lower_index = 16
        c3.create_pins(4)
        
        parser.current_definition = my_def

        return parser, c1, c2, c3

    def add_to_port_helper(self, tests, parser):
        #tests is a list of tuples that contain
        #name,left,right,direction,port,expected_lower,expected_width, expected_downto

        for entry in tests:
            name, left_index, right_index, direction, port, expected_lower, expected_width = entry
            expected_downto = port.is_downto #the port should not update it's downto once that has been decided.
            c = parser.create_or_update_port(name, left_index, right_index, direction)
            assert c is port, "the port is not the expected port, " + name
            assert len(c.pins) == expected_width, "the port is not the proper width, " + name
            assert c.lower_index == expected_lower, "the port does not have the proper lower index, " + name
            assert c.is_downto == expected_downto, "the port changed downto, " + name
            assert c.direction == direction, "the port is no longer "+ direction+ " it is now "+ c.direction+ " " + name


    def create_port_helper(self, tests, parser):
        #tests is a list of tuples that contain
        #name,left,right,direction,port,expected_lower,expected_width, expected_downto
        start_size = len(parser.current_definition.ports)
        count = 0
        for entry in tests:
            count += 1
            name, left_index, right_index, direction, port, expected_lower, expected_width, expected_downto = entry
            c = parser.create_or_update_port(name, left_index, right_index, direction)
            assert len(c.pins) == expected_width, "the port is not the proper width, " + name
            assert c.lower_index == expected_lower, "the port does not have the proper lower index, " + name
            assert c.is_downto == expected_downto
            assert c.direction == direction, "the port is no longer " + direction + " it is now " + c.direction+ " " + name
        assert len(parser.current_definition.ports) == count + start_size, "the wrong number of ports were added"


    def test_add_to_front_of_port(self):
        parser, c1, c2, c3 = self.init_port_creation()
        
        tests =\
            [("preexisting1", 15, 0, sdn.Port.Direction.IN, c1, 0, 16),\
            ("preexisting2", 0, -16, sdn.Port.Direction.OUT, c2, -16, 32),\
            ("preexisting3", 17, 0, sdn.Port.Direction.INOUT, c3, 0, 20)]
        self.add_to_port_helper(tests, parser)


    def test_add_to_back_of_port(self):
        parser, c1, c2, c3 = self.init_port_creation()
        
        tests =\
            [("preexisting1", 31, 16, sdn.Port.Direction.UNDEFINED, c1, 8, 24),\
            ("preexisting2", 31, 8, sdn.Port.Direction.IN, c2, 0, 32),\
            ("preexisting3", 63, 0, sdn.Port.Direction.OUT, c3, 0, 64)]
        self.add_to_port_helper(tests, parser)


    def test_create_port(self):
        parser, c1, c2, c3 = self.init_port_creation()
        
        tests =\
            [("new_port1", 31, 16, sdn.Port.Direction.IN, None, 16, 16, True),\
            ("new_port2", 0, -7, sdn.Port.Direction.INOUT, None, -7, 8, True),\
            ("new_port3", 63, 0, sdn.Port.Direction.UNDEFINED, None, 0, 64, True)]
        self.create_port_helper(tests, parser)


    def test_create_port_downto(self):
        parser, c1, c2, c3 = self.init_port_creation()
        
        tests =\
            [("new_port1", 7, 0, sdn.Port.Direction.OUT, None, 0, 8, True),\
            ("new_port2", 0, 7, sdn.Port.Direction.INOUT, None, 0, 8, False)]
        self.create_port_helper(tests, parser)

    def test_change_port_direction(self):
        pass #TODO not sure yet how I should deal with this.

    def test_dont_change_port(self):
        parser, c1, c2, c3 = self.init_port_creation()
        
        tests =\
            [("preexisting1", 15, 8, sdn.Port.Direction.OUT, c1, 8, 8),\
            ("preexisting2", 15, 0, sdn.Port.Direction.IN, c2, 0, 16),\
            ("preexisting3", 19, 16, sdn.Port.Direction.INOUT, c3, 16, 4)]
        self.add_to_port_helper(tests, parser)

    def test_single_index_port(self):
        parser, c1, c2, c3 = self.init_port_creation()
        tests_create =\
            [("new_port1", 31, None, sdn.Port.Direction.OUT, None, 31, 1, True),\
            ("new_port2", 0, None, sdn.Port.Direction.UNDEFINED, None, 0, 1, True),\
            ("new_port3", 63, None, sdn.Port.Direction.IN, None, 63, 1, True)]
        self.create_port_helper(tests_create, parser)

        parser, c1, c2, c3 = self.init_port_creation()
        tests_modify =\
            [("preexisting1", 0, None, sdn.Port.Direction.INOUT, c1, 0, 16),\
            ("preexisting2", 8, None, sdn.Port.Direction.OUT, c2, 0, 16),\
            ("preexisting3", 0, None, sdn.Port.Direction.IN, c3, 0, 20)]
        self.add_to_port_helper(tests_modify, parser)


    def test_change_port_downto(self):
        #i don't think the port downto-ness should ever change. this is tested in the add to port helper
        parser, c1, c2, c3 = self.init_port_creation()
        
        tests =\
            [("preexisting1", 15, 8, sdn.Port.Direction.UNDEFINED, c1, 8, 8),\
            ("preexisting1", 8, 15, sdn.Port.Direction.UNDEFINED, c1, 8, 8)]
        self.add_to_port_helper(tests, parser)

    def test_port_prepend_pins(self):
        parser, c1, c2, c3 = self.init_port_creation()
        c1_width = len(c1.pins)
        c1_lower = c1.lower_index
        c1_pins = set()
        for pin in c1.pins:
            c1_pins.add(pin)
        c2_width = len(c2.pins)
        c2_lower = c2.lower_index
        c2_pins = set()
        for pin in c2.pins:
            c2_pins.add(pin)
        c3_width = len(c3.pins)
        c3_lower = c3.lower_index
        c3_pins = set()
        for pin in c3.pins:
            c3_pins.add(pin)

        print(c1_pins)
        print()
        
        prepend_count = 8

        parser.prepend_pins(c1, prepend_count)
        print(c1_pins)
        print()
        print(c1.pins)
        print()
        assert len(c1.pins) == prepend_count + c1_width, "tried to add 8 pins to c1"
        assert c1.lower_index == c1_lower - prepend_count
        for pin in c1.pins[:prepend_count]:
            assert pin not in c1_pins, "a pin was in the wrong location for the prepend"
        for pin in c1.pins[prepend_count:]:
            assert pin in c1_pins, "a pin seems to have disappeared"

        parser.prepend_pins(c2, prepend_count)
        assert len(c2.pins) == prepend_count + c2_width, "tried to add 8 pins to c2"
        assert c2.lower_index == c2_lower - prepend_count
        for pin in c2.pins[:prepend_count]:
            assert pin not in c2_pins, "a pin was in the wrong location for the prepend"
        for pin in c2.pins[prepend_count:]:
            assert pin in c2_pins, "a pin seems to have disappeared"

        parser.prepend_pins(c3, prepend_count)
        assert len(c3.pins) == prepend_count + c3_width, "tried to add 8 pins to c3"
        assert c3.lower_index == c3_lower - prepend_count
        for pin in c3.pins[:prepend_count]:
            assert pin not in c3_pins, "a pin was in the wrong location for the prepend"
        for pin in c3.pins[prepend_count:]:
            assert pin in c3_pins, "a pin seems to have disappeared"

    def test_port_postpend_pins(self):
        parser, c1, c2, c3 = self.init_port_creation()
        c1_width = len(c1.pins)
        c1_lower = c1.lower_index
        c1_pins = set()
        for pin in c1.pins:
            c1_pins.add(pin)
        c2_width = len(c2.pins)
        c2_lower = c2.lower_index
        c2_pins = set()
        for pin in c2.pins:
            c2_pins.add(pin)
        c3_width = len(c3.pins)
        c3_lower = c3.lower_index
        c3_pins = set()
        for pin in c3.pins:
            c3_pins.add(pin)
        
        postpend_count = 8

        parser.postpend_pins(c1, postpend_count)
        assert len(c1.pins) == postpend_count + c1_width, "tried to add 8 pins to c1"
        assert c1.lower_index == c1_lower
        for pin in c1.pins[c1_width:]:
            assert pin not in c1_pins, "a pin was in the wrong location for the postpend"
        for pin in c1.pins[:c1_width]:
            assert pin in c1_pins, "a pin seems to have disappeared"

        parser.postpend_pins(c2, postpend_count)
        assert len(c2.pins) == postpend_count + c2_width, "tried to add 8 pins to c2"
        assert c2.lower_index == c2_lower
        for pin in c2.pins[c2_width:]:
            assert pin not in c2_pins, "a pin was in the wrong location for the postpend"
        for pin in c2.pins[:c2_width]:
            assert pin in c2_pins, "a pin seems to have disappeared"

        parser.postpend_pins(c3, postpend_count)
        assert len(c3.pins) == postpend_count + c3_width, "tried to add 8 pins to c3"
        assert c3.lower_index == c3_lower
        for pin in c3.pins[c3_width:]:
            assert pin not in c3_pins, "a pin was in the wrong location for the postpend"
        for pin in c3.pins[:c3_width]:
            assert pin in c3_pins, "a pin seems to have disappeared"

    def test_populate_new_port(self):
        parser = VerilogParser()
        
        test_data = [("test_name", 2, 0, sdn.Port.Direction.IN), ("\\escaped#$@_[213] ", 15,8, sdn.Port.Direction.OUT), ("some_name", 7,4,sdn.Port.Direction.INOUT)]

        for name, left, right, direction in test_data:
            port = sdn.Port()
            parser.populate_new_port(port,name,left,right,direction)
            assert port.name == name
            assert port.is_downto == (left >= right)
            assert port.direction == direction
            assert port.lower_index == min(left,right)

    def test_parse_port_declaration(self):
        parser = VerilogParser()
        count = 3
        names = ["i", "o", "io"]
        token_list = ["input", names[0], ";", "output", "[", "3", ":", "0", "]", names[1], ";", "inout", "wire", names[2], ";"]
        tokenizer = self.TestTokenizer(token_list)
        parser.tokenizer = tokenizer
        d = sdn.Definition()
        
        parser.current_definition = d

        for n in names:
            p = parser.create_or_update_port(n)
            c = parser.create_or_update_cable(n)
            c.wires[0].connect_pin(p.pins[0])

        for _ in range(count):
            parser.parse_port_declaration(dict())


    @unittest.expectedFailure
    def test_parse_port_declaration_failure(self):
        parser = VerilogParser()
        count = 3
        names = ["i", "o", "io"]
        token_list = ["input", names[0], ";", "output", "[", "3", ":", "0", "]", names[1], ";", "inout", "wire", names[2], ";"]
        tokenizer = self.TestTokenizer(token_list)
        parser.tokenizer = tokenizer
        d = sdn.Definition()
        
        parser.current_definition = d

        for n in names:
            parser.create_or_update_port(n)

        #the ports are in the definition but not connected this should fail

        for _ in range(count):
            parser.parse_port_declaration(properties = dict())

    def test_create_or_update_port_on_instance(self):
        parser = VerilogParser()
        parser.current_definition = sdn.Definition()
        parser.current_instance = parser.current_definition.create_child()
        parser.current_instance.reference = parser.blackbox_holder.get_blackbox("definition1")

        pins = parser.create_or_update_port_on_instance("port1", 1)

        assert parser.current_instance.reference.ports[0].name == "port1"
        assert len(parser.current_instance.reference.ports[0].pins) == 1
        assert len(parser.current_instance.pins) == 1
        assert len(pins) == 1

    ############################################
    ##assignment tests
    ############################################

    def test_create_new_assignment_instance(self):
        '''make sure the name, and width make sense
        also take advantage of the setup to make sure the definition returned is the same for the same width'''
        parser = VerilogParser()
        parser.netlist = sdn.Netlist()
        parser.current_definition = sdn.Definition()
        width = 4
        instance = parser.create_assignment_instance(width)
        assert len(instance.pins) == width * 2
        assert instance.name == "SDN_VERILOG_ASSIGNMENT_" + str(width) + "_" + str(0)
        assert instance.reference == parser.get_assignment_definition(width)

    def test_create_new_assignment_definition(self):
        '''make sure the names, width and acutal connectivity make sense'''
        parser = VerilogParser()
        parser.netlist = sdn.Netlist()
        parser.current_definition = sdn.Definition()
        width = 4
        definition = parser.get_assignment_definition(width)
        assert definition.name == "SDN_VERILOG_ASSIGNMENT_" + str(width)
        assert len(definition.ports) == 2
        assert len(definition.ports[0].pins) == width
        assert len(definition.ports[1].pins) == width
        assert len(definition.cables) == 1
        for i in range(width):
            assert definition.ports[0].pins[i] in definition.cables[0].wires[i].pins
            assert definition.ports[1].pins[i] in definition.cables[0].wires[i].pins


    def test_connect_assigned_wires(self):
        '''make sure the wires are actually connected to the instance and that the connectivity makes sense'''
        parser = VerilogParser()
        parser.netlist = sdn.Netlist()
        parser.current_definition = sdn.Definition()
        o_cable = parser.current_definition.create_cable(name = "c1")
        i_cable = parser.current_definition.create_cable(name = "c2")
        o_cable.create_wires(4)
        i_cable.create_wires(4)
        parser.connect_wires_for_assign(o_cable, 1, 0, i_cable, 3, 2)


    def test_names_of_instances_are_different(self):
        '''make sure the names of multiple instances of the same width are different'''
        parser = VerilogParser()
        parser.netlist = sdn.Netlist()
        parser.current_definition = sdn.Definition()
        width = 4
        inst1 = parser.create_assignment_instance(width)
        inst2 = parser.create_assignment_instance(width)
        assert inst1.name == "SDN_VERILOG_ASSIGNMENT_" + str(width) + "_" + str(0)
        assert inst2.name == "SDN_VERILOG_ASSIGNMENT_" + str(width) + "_" + str(1)

    def test_parse_assign(self):
        parser = VerilogParser()
        parser.netlist = sdn.Netlist()
        parser.current_definition = sdn.Definition()
        tokens = ["assign", "cable1", "=", "cable2", ";", "assign", "SR2", "[", "2", "]", "=", "\\<const0> ", ";"]
        tokenizer = self.TestTokenizer(tokens)
        parser.tokenizer = tokenizer
        c1, o_left, o_right, c2, i_left, i_right = parser.parse_assign()
        assert c1.name == "cable1"
        assert c2.name == "cable2"
        assert o_left == None
        assert o_right == None
        assert i_left == None
        assert i_right == None
        c1, o_left, o_right, c2, i_left, i_right = parser.parse_assign()
        assert c1.name == "SR2"
        assert c2.name == "\\<const0> "
        assert o_left == 2
        assert o_right == None
        assert i_left == None
        assert i_right == None

    ############################################
    ##Parse star parameters
    ############################################
    
    def test_parse_star_with_value(self):
        tokens = ['(', '*', 'key', '=', 'value', '*', ')']
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        
        stars = parser.parse_star_property()
        for k,v in stars.items():
            assert k == "key"
            assert v == "value"

    def test_parser_star_without_value(self):
        tokens = ['(', '*', 'key', '*', ')']
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        
        stars = parser.parse_star_property()
        for k,v in stars.items():
            assert k == "key"
            assert v is None

    def test_parser_star_list(self):
        '''example taken from a file that is not in our support files.
        the construct in question is referenced in xilinx documentation
        (* KEEP, DONT_TOUCH, BEL = "C6LUT" *)
        additionally:
        (*BEL="H6LUT",RLOC="X0Y0"*)
        and I presume that:
        (* KEEP, DONT_TOUCH *)
        would all be valid
        ''' 
        tokens = ["(", "*", "KEEP", ",", "DONT_TOUCH", ",", "BEL", "=", '"C6LUT"', "*", ")",\
            "(", "*", "BEL", "=", '"H6LUT"', ",", "RLOC", "=", '"X0Y0"', "*", ")",\
            "(", "*", "KEEP", ",", "DONT_TOUCH", "*", ")"]
        
        tokenizer = self.TestTokenizer(tokens)
        parser = VerilogParser()
        parser.tokenizer = tokenizer
        
        stars0 = parser.parse_star_property()
        stars1 = parser.parse_star_property()
        stars2 = parser.parse_star_property()

        assert "KEEP" in stars0
        assert stars0["KEEP"] == None
        assert "DONT_TOUCH" in stars0
        assert stars0["DONT_TOUCH"] == None
        assert "BEL" in stars0
        assert stars0["BEL"] == '"C6LUT"'

        assert "BEL" in stars1
        assert stars1["BEL"] == '"H6LUT"'
        assert "RLOC" in stars1
        assert stars1["RLOC"] == '"X0Y0"'

        assert "KEEP" in stars2
        assert stars2["KEEP"] == None
        assert "DONT_TOUCH" in stars2
        assert stars2["DONT_TOUCH"] == None
        


    ############################################
    ##test helpers
    ############################################

    def test_get_all_ports_from_wires(self):
        port1 = sdn.Port()
        port2 = sdn.Port()
        port1.create_pin()
        port2.create_pin()

        wire1 = sdn.Wire()
        wire2 = sdn.Wire()
        wire3 = sdn.Wire()

        wire1.connect_pin(port1.pins[0])
        wire2.connect_pin(port2.pins[0])

        groups = [[wire1, wire2, wire3], [wire1, wire2], [wire1], [wire3]]
        expected_results = [[port1, port2], [port1, port2], [port1], []]
        
        parser = VerilogParser()

        for i in range(len(groups)):
            g = groups[i]
            actual_results = parser.get_all_ports_from_wires(g)
            for r in actual_results:
                assert r in expected_results[i]
            for r in expected_results[i]:
                assert r in actual_results
        

    def test_get_wires_from_cable_helper(self):

        parser = VerilogParser()

        cable1 = sdn.Cable()
        cable1.create_wires(8)
        cable1.lower_index = 8

        cable2 = sdn.Cable()
        cable2.create_wires(8)
        cable2.lower_index = 0

        #add the cable.lower_index to all tests

        tests = [(7,0), (0, None), (0,0), (1,None), (2,4)]
        e11 = []
        e21 = []
        for i in reversed(range(0,8)):
            e11.append(cable1.wires[i])
            e21.append(cable2.wires[i])

        expected1 = [e11, [cable1.wires[0]], [cable1.wires[0]], [cable1.wires[1]], cable1.wires[2:5]]
        expected2 = [e21, [cable2.wires[0]], [cable2.wires[0]], [cable2.wires[1]], cable2.wires[2:5]]

        for i in range(len(tests)):
            left1 = tests[i][0] + cable1.lower_index
            if tests[i][1] != None:
                right1 = tests[i][1] + cable1.lower_index
            else:
                right1 = None
            left2 = tests[i][0] + cable2.lower_index
            if tests[i][1] != None:
                right2 = tests[i][1] + cable2.lower_index
            else:
                right2 = None
            wires1 = parser.get_wires_from_cable(cable1, left1, right1)
            wires2 = parser.get_wires_from_cable(cable2, left2, right2)

            for j in range(len(expected1[i])):
                w11 = wires1[j]
                w12 = expected1[i][j]
                assert w11 == w12, "the wires are not the same or not in the same order."
                w21 = wires2[j]
                w22 = expected2[i][j]
                assert w21 == w22, "the wires are not the same or not in the same order"



if __name__ == '__main__':
    unittest.main()



import unittest
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern


class TestPatterns(unittest.TestCase):
    def test_is_pattern_absolute(self):
        self.assertFalse(_is_pattern_absolute("MY_TEST", False, False))
        self.assertFalse(_is_pattern_absolute("MY_TEST", True, True))
        self.assertTrue(_is_pattern_absolute("MY_TEST", True, False))
        self.assertFalse(_is_pattern_absolute("*", True, False))
        self.assertFalse(_is_pattern_absolute(".*", True, True))

    def test_value_matches_pattern(self):
        self.assertTrue(_value_matches_pattern(None, "*", is_case=True, is_re=False))
        self.assertTrue(_value_matches_pattern(None, "*", is_case=False, is_re=False))
        self.assertTrue(_value_matches_pattern(None, ".*", is_case=True, is_re=True))
        self.assertTrue(_value_matches_pattern(None, ".*", is_case=False, is_re=True))

    def test_value_matches_pattern_bad_regex_expr(self):
        self.assertFalse(_value_matches_pattern(None, "*", is_case=False, is_re=True))




import unittest
import spydrnet as sdn
from spydrnet import get_libraries


class TestUtil(unittest.TestCase):
    def test_single_object(self):
        netlist = sdn.Netlist()
        library1 = netlist.create_library()
        library1.name = "work"
        library2 = netlist.create_library()
        library2.name = "hdi_primitives"
        library = next(get_libraries(netlist, "work"))
        self.assertEqual(library1, library)

    def test_multiple_objects(self):
        netlist1 = sdn.Netlist()
        library1 = netlist1.create_library()
        netlist2 = sdn.Netlist()
        library2 = netlist2.create_library()
        get_libraries([netlist1, netlist2])




from spydrnet.ir.first_class_element import FirstClassElement
from spydrnet.ir.definition import Definition
from spydrnet.ir.views.listview import ListView
from spydrnet.global_state import global_callback
from spydrnet.global_state.global_callback import _call_create_library
from copy import deepcopy, copy, error


class Library(FirstClassElement):
    """
    Represents a library object.

    Contains a pointer to parent netlist and definitions.
    """
    __slots__ = ['_netlist', '_definitions']

    def __init__(self, name = None, properties = None):
        """
        creates an empty object of type Library

        parameters
        ----------

        name - (str) the name of this instance
        properties - (dict) the dictionary which holds the properties
        """
        super().__init__()
        self._netlist = None
        self._definitions = list()
        _call_create_library(self)
        if name!= None:
            self.name = name
        if properties != None:
            assert isinstance(properties, dict), "properties must be a dictionary"
            for key in properties:
                self[key] = properties[key]

    @property
    def netlist(self):
        """
        get the netlist that contains this library
        """
        return self._netlist

    @property
    def definitions(self):
        """
        return a list of all the definitions that are included in this library
        """
        return ListView(self._definitions)

    @definitions.setter
    def definitions(self, value):
        """
        set the definitions to a new reordered set of definitions. This function cannot be used to add or remove
        definitions

        Parameters
        ----------

        value - (List containing Definition type objects) The reordered list
        """
        value_list = list(value)
        value_set = set(value_list)
        assert len(value_list) == len(value_set) and set(self._definitions) == value_set, \
            "Set of values do not match, this function can only reorder values, values must be unique"
        self._definitions = value_list

    def create_definition(self):
        """
        create a definition, add it to the library, and return the definition
        """
        definition = Definition()
        self.add_definition(definition)
        return definition

    def add_definition(self, definition, position=None):
        """
        add an existing definition to the library. The definition must not belong to a library including this one.

        parameters
        ----------

        definition - (Definition) the defintion to add to the library

        position - (int, default None) the index in the library list at which to add the definition

        """
        assert definition.library is not self, "Definition already included in library"
        assert definition.library is None, "Definition already belongs to a different library"
        global_callback._call_library_add_definition(self, definition)
        if position is not None:
            self._definitions.insert(position, definition)
        else:
            self._definitions.append(definition)
        definition._library = self

    def remove_definition(self, definition):
        """
        remove the given definition from the library

        parameters
        ----------

        definition - (Definition) the definition to be removed
        """
        assert definition.library == self, "definition is not included in library"
        self._remove_definition(definition)
        self._definitions.remove(definition)

    def remove_definitions_from(self, definitions):
        """
        remove a set of definitions from the library. all definitions provided must be in the library

        parameters
        ----------

        definitions - (Set of Definition type objects) the definitions to be removed
        """
        if isinstance(definitions, set):
            excluded_definitions = definitions
        else:
            excluded_definitions = set(definitions)
        assert all(x.library == self for x in excluded_definitions), "Some definitions to remove are not included in " \
                                                                     "the library "
        included_definitions = list()
        for definition in self._definitions:
            if definition not in excluded_definitions:
                included_definitions.append(definition)
            else:
                self._remove_definition(definition)
        self._definitions = included_definitions

    def _remove_definition(self, definition):
        """
        internal function to dissociate a definition from the library
        """
        global_callback._call_library_remove_definition(self, definition)
        definition._library = None


    def _clone_rip_and_replace(self, memo):
        '''remove from its current environment and place it into the new cloned environment with references held in the memo dictionary'''
        pass #this function will need to call rip and replace in library on each of the definitions when called from the netlist.
        for definition in self._definitions:
            definition._clone_rip_and_replace(memo)

    def _clone_rip(self, memo):
        '''remove from its current environmnet. This will remove all pin pointers and create a floating stand alone instance.'''   
        # references lists of definitions need to be vacated except those that were cloned.
        for definition in self._definitions:
            new_references = set()
            for ref in definition._references:
                if ref in memo.values():
                    new_references.add(ref)
            for instance in definition._children:
                instance._reference._references.add(instance)
                
            definition._references = new_references


    def _clone(self, memo):
        '''not api safe clone function
        clone leaving all references in tact.
        the element can then either be ripped or ripped and replaced'''
        assert self not in memo, "the object should not have been copied twice in this pass"
        c = Library()
        memo[self] = c
        c._netlist = None
        c._data = deepcopy(self._data)
        
        new_definitions = list()
        for definition in self._definitions:
            new_definitions.append(definition._clone(memo))
        c._definitions = new_definitions

        for definition in c._definitions:
            definition._library = c
            definition._clone_rip_and_replace(memo)
        return c

    def clone(self):
        '''
        Clone the library in an api safe manner.
        The following describes the structure of the returned object:
         * the instances that pointed to reference definitions within the library will have updated references
         * the instances that pointed to reference definitions outside the library will maintain their definitions
         * the references lists (of definitions) both inside and outsde the library will be updated to reflect the change
         * all definitions are cloned within the library.
         '''
        memo = dict()
        c = self._clone(memo)
        c._clone_rip(memo)
        return c



# Copyright 2021 please see the license
# Author Dallin Skouson

from functools import partial
import re
import zipfile
import io
import os
import spydrnet.parsers.verilog.verilog_tokens as vt
from spydrnet.parsers.verilog.verilog_token_factory import TokenFactory


class VerilogTokenizer:
    @staticmethod
    def from_stream(stream):
        tokenizer = VerilogTokenizer(stream)
        return tokenizer

    @staticmethod
    def from_string(string):
        string_stream = io.StringIO(string)
        tokenizer = VerilogTokenizer(string_stream)
        return tokenizer

    @staticmethod
    def from_filename(filename):
        tokenizer = VerilogTokenizer(filename)
        return tokenizer

    def __init__(self, input_source):
        self.token = None
        self.next_token = None
        self.line_number = 0

        if isinstance(input_source, str):
            if zipfile.is_zipfile(input_source):
                zip = zipfile.ZipFile(input_source)
                filename = os.path.basename(input_source)
                filename = filename[:filename.rindex(".")]
                stream = zip.open(filename)
                stream = io.TextIOWrapper(stream)
                self.input_stream = stream
            else:
                self.input_stream = open(input_source, 'r')
        else:
            if isinstance(input_source, io.TextIOBase) is False:
                self.input_stream = io.TextIOWrapper(input_source)
            else:
                self.input_stream = input_source

        self.generator = self.generate_tokens()

    def __del__(self):
        if hasattr(self, "input_stream"):
            self.close()

    def has_next(self):
        try:
            self.peek()
            return True
        except StopIteration:
            return False

    def next(self):
        if self.next_token is not None:
            self.token = self.next_token
            self.next_token = None
        else:
            self.token = next(self.generator)
        return self.token

    def peek(self):
        if self.next_token is not None:
            return self.next_token
        else:
            token = next(self.generator)
            while len(token) >= 2 and (token[0:2] == vt.OPEN_LINE_COMMENT
                                       or token[0:2] == vt.OPEN_BLOCK_COMMENT):
                token = next(self.generator)
            self.next_token = token
            return self.next_token

    def generate_tokens(self):
        '''give independent tokens from the token factory'''

        try:
            self.line_number = 1
            tf = TokenFactory()
            for buffer in iter(partial(self.input_stream.read, 32768), ""):
                for ch in buffer:
                    if ch == vt.NEW_LINE:
                        self.line_number += 1
                    result = tf.add_character(ch)
                    if result is not None:
                        yield result
        finally:
            self.input_stream.close()

        # if the input doesn't end in white space there will be one token left in the token factory try and get it.

        result = tf.flush()
        if result != None:
            yield result

    def close(self):
        if self.input_stream:
            self.input_stream.close()



from spydrnet.ir import Netlist, Library, Definition


class DefaultNamespace:
    @classmethod
    def is_compliant(cls, element):
        """
        is_compliant means that the naming of the element is a valid name and that there are no namespace conflicts
        among its children
        :param element:
        :return:
        """
        if cls.is_name_of_element_valid(element) and cls.no_name_conflicts(element):
            return True
        return False

    @classmethod
    def is_name_of_element_valid(cls, element):
        """
        All names are valid in the default namespace. Always returns true.
        :param element:
        :return: True
        """
        return True

    @classmethod
    def is_name_valid(cls, key, value):
        return True

    @classmethod
    def no_name_conflicts(cls, element):
        """
        Check to see if there are any name conflicts among the children
        :param element:
        :return:
        """
        namespace = set()
        if isinstance(element, Netlist):
            for library in element.libraries:
                if ".NAME" in library:
                    name = library[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
        elif isinstance(element, Library):
            for definition in element.definitions:
                if ".NAME" in definition:
                    name = definition[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
        elif isinstance(element, Definition):
            for port in element.ports:
                if ".NAME" in port:
                    name = port[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
            namespace.clear()
            for cable in element.cables:
                if ".NAME" in cable:
                    name = cable[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
            namespace.clear()
            for instance in element.children:
                if ".NAME" in instance:
                    name = instance[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
        return True

    @classmethod
    def needs_namespace(cls, element):
        if isinstance(element, (Netlist, Library, Definition)):
            return True
        return False

    def __init__(self):
        self.namespaces = dict()

    def no_conflict(self, element, key, value):
        if key != ".NAME":
            return True
        element_type = type(element)
        if element_type in self.namespaces:
            namespace = self.namespaces[element_type]
            if value in namespace:
                if namespace[value] != element:
                    return False
        return True

    def update(self, element, key, value):
        if key == ".NAME":
            element_type = type(element)
            if element_type not in self.namespaces:
                self.namespaces[element_type] = dict()
            namespace = self.namespaces[element_type]
            if ".NAME" in element:
                old_name = element[".NAME"]
                if old_name in namespace:
                    del namespace[old_name]
            namespace[value] = element

    def remove(self, element, key):
        if key == ".NAME":
            element_type = type(element)
            if element_type in self.namespaces:
                namespace = self.namespaces[element_type]
                if ".NAME" in element:
                    old_name = element[".NAME"]
                    if old_name in namespace:
                        del namespace[old_name]

    def lookup(self, element_type, key, value):
        if key == ".NAME":
            if element_type in self.namespaces:
                namespace = self.namespaces[element_type]
                return namespace.get(value, None)



import re


class EdififyNames:
    """
    Handles the renaming of objects from other languages to EDIF. Emphasis
    is currently given to verilog.

    The following are differences between the namespaces of EDIF and Verilog.

    EDIF identifiers:

     * less than 256 characters in length (255 max)
     * a-z,A-Z,_,0-9 are valid characters
     * if non alphabetic character is first character prefix with &
     * case insensitive

    Verilog identifiers:

     * 1024 characters or less in length (1024 max)
     * a-z,A-Z,_,0-9 are valid characters. first character not digit (when unescpaed)
     * case sensitive
     * identifier starting with backslash is escaped
     * escaped identifiers can contain any ascii character except white space

    This class applys an approach similar to vivado to create the rename

     * shorten identifiers of too great a length then postfix with a unique id
     * replace invalid characters with _
     * conflicts when case insensitivity is used are postfixed with a unique id
     * names that end up matching other names are postfixed with a unique id

    Examples
    --------
    >>> from spydrnet.composers.edif.edifify_names import EdififyNames
    >>> import spydrnet as sdn
    >>> ed = EdififyNames()
    >>> i = sdn.Instance()
    >>> i.name = "*this_is+an$*`id[0:3]"
    >>> l = [i]
    >>> valid_identifier = ed.make_valid(i,l)

    The valid_identifier should be the following:

    &_this_is_an___in_0_3_

    >>> import spydrnet as sdn
    >>> i = sdn.Instance()
    >>> i.name = 'name'
    >>> i2 = sdn.Instance()
    >>> i2.name = 'name_sdn_1_'
    >>> i3 = sdn.Instance()
    >>> i3.name = 'name'
    >>> my_list = [i,i2,i3]
    >>> ed.make_valid(i3, my_list)

    The output should be the following:

    name_sdn_2_

    """

    def __init__(self):
        self.valid = set()
        self.non_alpha = set()
        self.name_length_target = 100

    def _length_good(self, identifier):
        """returns a boolean indicating whether or not the indentifier fits the 256 character limit"""

        return len(identifier) < self.name_length_target

    def _length_fix(self, identifier):
        """returns the name with the fixed length of 256 characters if the limit is exceeded"""
        if not self._length_good(identifier):
            pattern = re.compile('_sdn_[0-9]+_$')
            r = pattern.search(identifier)
            if r is None:
                return identifier[:self.name_length_target]
            else:
                return identifier[:self.name_length_target - (r.end() - r.start())] + identifier[r.start():]
        else:
            return identifier

    def _characters_good(self, identifier):
        """Check if the characters meet the edif standards

        returns whether the string only contain numbers, characters and '-'
        """
        if not identifier[0].isalpha():
            return False
        for i in range(0, len(identifier)):
            if not identifier[i].isalnum() and identifier[i] != '-':
                return False
        return True

    def _characters_fix(self, identifier):
        """fix the characters so that it meets edif standards

        Add a '&' if the first character is not alphabetic
        replaces all the characters to '-' if it is not valid characters
        """
        if not self._characters_good(identifier):
            starting_index = 0
            if not identifier[0].isalpha():
                identifier = '&' + identifier[:]
                starting_index = 1

            for i in range(starting_index, len(identifier)):
                if not identifier[i].isalnum():
                    identifier = identifier[: i] + '_' + identifier[i+1:]
        identifier = self._length_fix(identifier)
        return identifier

    def _conflicts_good(self, obj, identifier, objects):
        for element in objects:
            if element == obj:
                continue
            if element.name == identifier or ("EDIF.identifier" in element.data and element["EDIF.identifier"] == identifier):
                return False
        return True

    def _conflicts_fix(self, obj, identifier, objects):
        identifier_lower = identifier.lower()
        if not self._conflicts_good(obj, identifier_lower, objects):
            pattern = re.compile('_sdn_[0-9]+_$')
            r = pattern.search(identifier_lower)
            if r is None:
                identifier_lower = identifier_lower + '_sdn_1_'
            else:
                # get the number out of the string
                num = int(
                    re.search(r'\d+', identifier_lower[r.start():]).group())
                identifier_lower = identifier_lower[:r.start(
                )+5] + str(num + 1) + '_'
            identifier_lower = self._length_fix(identifier_lower)
            identifier_lower = self._conflicts_fix(
                obj, identifier_lower, objects)
            identifier = identifier_lower
        return identifier

    def is_valid_identifier(self, identifier):
        """
        check if the identifier is valid in the namespace that is set. Will also
        check to make sure the identifer is valid.
        """
        if self._length_good(identifier) is False:
            return False

        if self._characters_good(identifier) is False:
            return False
        return True

    def make_valid(self, obj, objects):
        """
        make a compliant identifier based on the identifier given.
        returns the identifier if no change is needed.
        """
        identifier = obj.name
        identifier = self._length_fix(identifier)
        identifier = self._characters_fix(identifier)
        identifier = self._conflicts_fix(obj, identifier, objects)

        return identifier



from spydrnet.ir import Netlist, Library, Definition, Port, Cable, Instance

_registered_lookups = dict()
_registered_hierarchical_lookup = None


def register_lookup(key, func):
    if key in _registered_lookups:
        raise ValueError(
            "Cannot register a fast lookup under the key {}, lookup already registered.")
    else:
        _registered_lookups[key] = func


def deregister_lookup(key):
    if key in _registered_lookups:
        del _registered_lookups[key]


def lookup(parent, element_type, key, value):
    if key in _registered_lookups:
        registered_lookup = _registered_lookups[key]
        return registered_lookup(parent, element_type, key, value)
    else:
        if isinstance(parent, Netlist):
            if element_type is Library:
                for library in parent.libraries:
                    if key in library and value == library[key]:
                        return library
        elif isinstance(parent, Library):
            if element_type is Definition:
                for definition in parent.definitions:
                    if key in definition and value == definition[key]:
                        return definition
        elif isinstance(parent, Definition):
            if element_type is Port:
                for port in parent.ports:
                    if key in port and value == port[key]:
                        return port
            elif element_type is Cable:
                for cable in parent.cables:
                    if key in cable and value == cable[key]:
                        return cable
            elif element_type is Instance:
                for instance in parent.children:
                    if key in instance and value == instance[key]:
                        return instance
    return None




import unittest
import spydrnet as sdn
from spydrnet.ir import *
from spydrnet.flatten import flatten
from spydrnet.uniquify import uniquify

class TestUniquify(unittest.TestCase):
    def create_netlist(self):
        nl = Netlist()
        lib = nl.create_library()
        d1 = lib.create_definition()
        d1.name = ("d1")
        d2 = lib.create_definition()
        d2.name = ("d2")
        d3 = lib.create_definition()
        d3.name = ("d3")
        i11 = d1.create_child()
        i11.name = ("i11")
        i12 = d1.create_child()
        i12.name = ("i12")
        i13 = d1.create_child()
        i13.name = ("i13")
        i24 = d2.create_child()
        i24.name = ("i24")
        i25 = d2.create_child()
        i25.name = ("i25")
        i26 = d2.create_child()
        i26.name = ("i26")
        i27 = d2.create_child()
        d4 = lib.create_definition()
        d4.name = ("d4")
        i27.name = ("i27")
        i38 = d3.create_child()
        i38.name = ("i38")
        i39 = d3.create_child()
        i39.name = ("i39")
        i11.reference = d3
        i12.reference = d3
        i13.reference = d3
        i24.reference = d1
        i25.reference = d1
        i26.reference = d3
        i27.reference = d3
        i38.reference = d4
        i39.reference = d4
        nl.top_instance = Instance()
        nl.top_instance.reference = d2
        nl.top_instance.name = ("top_instance")
        return nl

    def create_mem_bus(self,top, proc, mem):
        mem_width = 10
        mem_write_out = proc.create_port()
        mem_write_out.name = "mem_write_out"
        mem_write_out.create_pins(mem_width)
        mem_read_in = proc.create_port()
        mem_read_in.name = "mem_read_in"
        mem_read_in.create_pins(mem_width)

        mem_write_in = mem.create_port()
        mem_write_in.name = "mem_write_in"
        mem_write_in.create_pins(mem_width)
        mem_read_out = mem.create_port()
        mem_read_out.name = "mem_read_out"
        mem_read_out.create_pins(mem_width)

        mem_write = top.create_cable()
        mem_write.name = "mem_write"
        mem_write.create_wires(mem_width)

        mem_read = top.create_cable()
        mem_read.name = "mem_read"
        mem_read.create_wires(mem_width)

        proc_i = top.create_child()
        proc_i.name = "proc_i"
        proc_i.reference = proc
        mem_i = top.create_child()
        mem_i.name = "mem_i"
        mem_i.reference = mem

        for i in range(mem_width):
            mem_write.wires[i].connect_pin(mem_i.pins[mem_write_in.pins[i]])
            mem_write.wires[i].connect_pin(proc_i.pins[mem_write_out.pins[i]])
            mem_read.wires[i].connect_pin(mem_i.pins[mem_read_out.pins[i]])
            mem_read.wires[i].connect_pin(proc_i.pins[mem_read_in.pins[i]])

        return mem_write_out, mem_read_in


    def create_proc(self, proc, mem_write, mem_read, core, cachel2):
        mem_width = 10
        core_i1 = proc.create_child()
        core_i2 = proc.create_child()
        core_i1.reference = core
        core_i2.reference = core
        core_i1.name = "core_i1"
        core_i2.name = "core_i2"

        mem_p = core.create_port()
        mem_p.create_pins(mem_width)
        mem_p.name = "mem_p"

        cachel2_i = proc.create_child()
        cachel2_i.reference = cachel2
        cachel2_i.name = "cachel2_i"
        
        mem_a_in  = cachel2.create_port()
        mem_a_out = cachel2.create_port()
        mem_b_in  = cachel2.create_port()
        mem_b_out = cachel2.create_port()
        
        
        mem_a_in.create_pins(mem_width)
        mem_a_out.create_pins(mem_width)
        mem_b_in.create_pins(mem_width)
        mem_b_out.create_pins(mem_width)

        mem_a_in.name = "mem_a_in"
        mem_b_out.name = "mem_a_out"
        mem_b_out.name = "mem_b_out"
        mem_b_in.name = "mem_b_in"
        
        
        ca1 = proc.create_cable()
        ca2 = proc.create_cable()
        cb1 = proc.create_cable()
        cb2 = proc.create_cable()

        ca1.create_wires(mem_width)
        ca2.create_wires(mem_width)
        cb1.create_wires(mem_width)
        cb2.create_wires(mem_width)

        ca1.name = "ca1"
        ca2.name = "ca2"
        cb1.name = "cb1"
        cb2.name = "cb2"

        for i in range(mem_width):
           ca1.wires[i].connect_pin(core_i1.pins[mem_p.pins[i]])
           ca1.wires[i].connect_pin(cachel2_i.pins[mem_a_in.pins[i]])
           ca2.wires[i].connect_pin(mem_read.pins[i])
           ca2.wires[i].connect_pin(cachel2_i.pins[mem_a_out.pins[i]])
           cb1.wires[i].connect_pin(core_i2.pins[mem_p.pins[i]])
           cb1.wires[i].connect_pin(cachel2_i.pins[mem_b_in.pins[i]])
           cb2.wires[i].connect_pin(mem_write.pins[i])
           cb2.wires[i].connect_pin(cachel2_i.pins[mem_b_out.pins[i]])

        return mem_p

    def create_core(self, core, mem_p, alu, reg_file):
        mem_width = 10
        alu_i = core.create_child()
        reg_file_i = core.create_child()
        
        alu_i.reference = alu
        alu_i.name = "alu_i"
        reg_file_i.reference = reg_file
        reg_file_i.name = "reg_file_i"

        alu_p = alu.create_port()
        reg_p = reg_file.create_port()

        alu_p.create_pins(mem_width)
        reg_p.create_pins(mem_width)
        alu_p.name = "alu_p"
        reg_p.name = "reg_p"

        mem_cable = core.create_cable()

        mem_cable.create_wires(mem_width)
        mem_cable.name = "mem_cable"

        for i in range(mem_width):
            mem_cable.wires[i].connect_pin(reg_file_i.pins[reg_p.pins[i]])
            mem_cable.wires[i].connect_pin(alu_i.pins[alu_p.pins[i]])
            mem_cable.wires[i].connect_pin(mem_p.pins[i])

    def create_netlist_with_wires(self):
        nl = Netlist()
        lib = nl.create_library()
        top = lib.create_definition()
        top.name = "top"
        proc = lib.create_definition()
        proc.name = "proc"
        core = lib.create_definition()
        core.name = "core"
        alu = lib.create_definition()
        alu.name = "alu"
        mem = lib.create_definition()
        mem.name = "mem"
        cachel2 = lib.create_definition()
        cachel2.name = "cacheL2"
        reg_file = lib.create_definition()
        reg_file.name = "reg_file"
        nl.top_instance = Instance()
        nl.top_instance.reference = top

        
        mem_write, mem_read= self.create_mem_bus(top, proc, mem)

        core_mem = self.create_proc(proc, mem_write, mem_read, core, cachel2)

        self.create_core(core, core_mem, alu, reg_file)

        nl.top_instance.name = ("top_instance")
        
        #self.simple_recursive_netlist_visualizer(nl)

        return nl



    # def simple_recursive_netlist_visualizer(self, netlist):
    #     #TODO put this code somewhere where people can use it to visualize simple netlists
    #     top_instance = netlist.top_instance
    #     #should look something like this:
    #     #top
    #     #   child1
    #     #       child1.child
    #     #   child2
    #     #       child2.child
    #     def recurse(instance, depth):
    #         s = depth * "\t"
    #         print(s, instance.name, "(", instance.reference.name, ")")
    #         for c in instance.reference.children:  
    #             recurse(c, depth + 1)
        
    #     recurse(top_instance, 0)

    def simple_cable_connection_visualizer(self,netlist):
        top_instance = netlist.top_instance
        #display all the cables in the top instance with all of their connections next to it
        def recur(instance):
            print(instance.name)
            for c in instance.reference.cables:
                print("\t(cable): ", c.name)
                if len(c.wires) > 0:
                    w = c.wires[0]
                    for p in w.pins:
                        if isinstance(p,OuterPin):
                            print("\t\t(instance): ", p.instance.name)
                        else:
                            print("\t\t(port):     ", p.port.name)
            for i in instance.reference.children:
                recur(i)

        recur(top_instance)


    def is_flat(self,nl):
        ti = nl.top_instance
        td = ti.reference
        for i in td.children:
            if not i.reference.is_leaf():
                return False
        return True



    def test_flatten_instances(self):
        nl = self.create_netlist()
        uniquify(nl)
        flatten(nl)
        assert self.is_flat(nl)


    def test_flatten_cables(self):
        nl = self.create_netlist_with_wires()
        self.simple_cable_connection_visualizer(nl)
        
        uniquify(nl)
        flatten(nl)
        self.simple_cable_connection_visualizer(nl)
        
        assert self.is_flat(nl) #might be nice to add some tests for the connections here.


from spydrnet.ir import Element, FirstClassElement, InnerPin, OuterPin, Wire, Netlist, Library, Definition, Port, \
    Cable, Instance
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern


def get_netlists(obj, *args, **kwargs):
    """
    get_netlists(obj, ...)

    Get netlists *within* an object.
    
    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_libraries(netlist, ...)` would
        return all of the libraries associated with the provided netlist that match the additional criteria.
    
    patterns : str, Iterable - optional, positional or named, (default: wildcard)
        The search patterns. Patterns can be a single string or an Iterable collection of strings. Patterns can be
        absolute or they can contain wildcards or regular expressions. If `patterns` is not provided, then it defaults
        to a wildcard. Patterns are queried against the object property value stored under a specified key. Fast lookups
        are only attempted on absolute patterns that are not regular expressions and contain no wildcards.
    key : str, optional, (default: ".NAME")
        This is the key that controls which value is being searched.
    is_case : bool - optional, named, (default: True)
        Specify if patterns should be treated as case sensitive. Only applies to patterns. Does not alter fast lookup
        behavior (if namespace policy uses case insensitive indexing, this parameter will not prevent a fast lookup
        from returning a matching object even if the case is not an exact match).
    is_re: bool - optional, named, (default: False)
        Specify if patterns are regular expressions. If `False`, a pattern can still contain `*` and `?` wildcards. A
        `*` matches zero or more characters. A `?` matches upto a single character.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    netlists : generator
        A generator associated with a particular object
    
    """
    # Check argument list
    if len(args) == 1 and 'patterns' in kwargs:
        raise TypeError("get_netlists() got multiple values for argument 'patterns'")
    if len(args) > 1 or any(x not in {'patterns', 'key', 'filter', 'is_case', 'is_re'} for x in kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    filter_func = kwargs.get('filter', lambda x: True)
    is_case = kwargs.get('is_case', True)
    is_re = kwargs.get('is_re', False)
    patterns = args[0] if len(args) == 1 else kwargs.get('patterns', ".*" if is_re else "*")
    key = kwargs.get('key', ".NAME")

    if isinstance(obj, (FirstClassElement, InnerPin, OuterPin, Wire)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (Element, HRef)) for x in object_collection) is False:
        raise TypeError("get_netlists() supports netlist elements and hierarchical references or a collection of "
                        "theses as the object searched, unsupported object provided")

    if isinstance(patterns, str):
        patterns = (patterns,)

    return _get_netlists(object_collection, patterns, key, is_case, is_re, filter_func)


def _get_netlists(object_collection, patterns, key, is_case, is_re, filter_func):
    for result in filter(filter_func, _get_netlists_raw(object_collection, patterns, key, is_case, is_re)):
        yield result


def _get_netlists_raw(object_collection, patterns, key, is_case, is_re):
    found = set()
    namemap = dict()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Netlist):
            if obj not in found:
                found.add(obj)
                name = obj.get(key, None)
                if name not in namemap:
                    namemap[name] = list()
                namemap[name].append(obj)
        elif isinstance(obj, Library):
            netlist = obj.netlist
            if netlist:
                object_collection.append(netlist)
        elif isinstance(obj, Definition):
            library = obj.library
            if library:
                object_collection.append(library)
        elif isinstance(obj, Instance):
            reference = obj.reference
            if reference:
                object_collection.append(reference)
        elif isinstance(obj, Port):
            definition = obj.definition
            if definition:
                object_collection.append(definition)
        elif isinstance(obj, InnerPin):
            port = obj.port
            if port:
                object_collection.append(port)
        elif isinstance(obj, OuterPin):
            inner_pin = obj.inner_pin
            if inner_pin:
                object_collection.append(inner_pin)
        elif isinstance(obj, Cable):
            definition = obj.definition
            if definition:
                object_collection.append(definition)
        elif isinstance(obj, Wire):
            cable = obj.cable
            if cable:
                object_collection.append(cable)
        elif isinstance(obj, HRef):
            if obj.is_valid:
                item = obj.item
                if item:
                    object_collection.append(item)

    for pattern in patterns:
        pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
        if pattern_is_absolute:
            if pattern in namemap:
                result = namemap[pattern]
                for netlist in result:
                    if netlist in found:
                        found.remove(netlist)
                        yield netlist
        else:
            yielded = set()
            for netlist in found:
                value = netlist[key] if key in netlist else ''
                if _value_matches_pattern(value, pattern, is_case, is_re):
                    yielded.add(netlist)
                    yield netlist
            found -= yielded



import unittest
import os
import tempfile
import glob
import shutil

from spydrnet.compare.compare_netlists import Comparer
import spydrnet as sdn


class TestCompareNetlists(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.dir_of_edif_netlists = os.path.join(sdn.base_dir, "support_files", "EDIF_netlists")
        cls.dir_of_verilog_netlists = os.path.join(sdn.base_dir, "support_files", "verilog_netlists")
        cls.edif_files = sorted(glob.glob(os.path.join(cls.dir_of_edif_netlists, "*.edf.zip")), key=os.path.getsize)
        cls.verilog_files = sorted(glob.glob(os.path.join(cls.dir_of_verilog_netlists, "*.v.zip")), key = os.path.getsize)

    @unittest.skip("Test takes a long time right now.")
    def test_large_edif(self):
        for ii, filename in enumerate(self.edif_files):
            if os.path.getsize(filename) <= 1024 * 10:
                continue
            self.compare_parser_and_composer(filename, ii, "edf")

    def test_small_edif(self):
        for ii, filename in enumerate(self.edif_files):
            if os.path.getsize(filename) > 1024 * 10:
                continue
            self.compare_parser_and_composer(filename, ii, "edf")

    @unittest.skip("Test takes a long time right now.")
    def test_large_verilog(self):
        for ii, filename in enumerate(self.verilog_files):
            if os.path.getsize(filename) <= 1024 * 10:
                continue
            self.compare_parser_and_composer(filename, ii, "v")
        #assert False

    def test_small_verilog(self):
        for ii, filename in enumerate(self.verilog_files):
            if os.path.getsize(filename) > 1024 * 10:
                continue
            self.compare_parser_and_composer(filename, ii, "v")
        #assert False

    def compare_parser_and_composer(self,filename, ii, target_format_extension = None):
        with self.subTest(i=ii):
            if os.path.exists("temp"):
                shutil.rmtree("temp")
            print(filename)
            with tempfile.TemporaryDirectory() as tempdirname:
                try:
                    orig_netlist = sdn.parse(filename)
                    basename_without_final_ext = os.path.splitext(os.path.basename(filename))[0]
                    if target_format_extension is None:
                        composer_filename = os.path.join(tempdirname, basename_without_final_ext)
                    else:
                        composer_filename = os.path.join(tempdirname, basename_without_final_ext + "." + target_format_extension)
                    orig_netlist.compose(composer_filename)
                    print(composer_filename)
                    composer_netlist = sdn.parse(composer_filename)
                    comparer = Comparer(orig_netlist, composer_netlist)
                    comparer.run()
                except Exception as e:
                    shutil.copytree(tempdirname, "temp")
                    raise e

    @unittest.skip("Test takes a long time right now.")
    def test_large_verilog_to_edif(self):
        for ii, filename in enumerate(self.verilog_files):
            if os.path.getsize(filename) <= 1024 * 10:
                continue
            self.compare_parser_and_composer(filename, ii, "edf")
        #assert False

    @unittest.skip("currently not working properly for the number of cables on some examples please use with caution")
    def test_small_verilog_to_edif(self):
        for ii, filename in enumerate(self.verilog_files):
            if os.path.getsize(filename) > 1024 * 10:
                continue
            self.compare_parser_and_composer(filename, ii, "edf")
        #assert False
        
    def test_empty_netlists(self):
        nl1 = sdn.ir.Netlist()
        nl2 = sdn.ir.Netlist()
        comp = Comparer(nl1, nl2)
        comp.compare()


if __name__ == '__main__':
    unittest.main()



import unittest
import spydrnet as sdn


class TestGetPins(unittest.TestCase):
    def test_parameter_checking(self):
        definition = sdn.Definition()
        cable = definition.create_cable()
        cable.name = "MY_CABLE"
        self.assertRaises(TypeError, sdn.get_pins, definition, "MY_CABLE")
        self.assertRaises(TypeError, sdn.get_pins, definition, unsupported_keyword=None)
        self.assertRaises(TypeError, sdn.get_pins, definition, selection=None)
        self.assertRaises(TypeError, sdn.get_pins, None)
        self.assertRaises(TypeError, sdn.get_pins, [None, definition])

    def test_of_port_inside_and_outside(self):
        definition = sdn.Definition()
        port = definition.create_port()
        pin = port.create_pin()
        instance = sdn.Instance()
        instance.reference = definition

        search = next(sdn.get_pins(instance))
        self.assertEqual(search, pin)

        search = next(sdn.get_pins(instance, selection="OUTSIDE"))
        self.assertEqual(search, instance.pins[pin])

    def test_of_inner_pin_outside(self):
        definition = sdn.Definition()
        port = definition.create_port()
        pin = port.create_pin()
        instance = sdn.Instance()
        instance.reference = definition

        search = next(sdn.get_pins(pin, selection="OUTSIDE"))
        self.assertEqual(search, instance.pins[pin])

    def test_of_definition(self):
        definition = sdn.Definition()
        port = definition.create_port()
        pin = port.create_pin()
        instance = sdn.Instance()
        instance.reference = definition

        search = next(sdn.get_pins(definition))
        self.assertEqual(search, pin)

    def test_of_href_to_cable(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        port = definition.create_port()
        pin = port.create_pin()
        cable = definition.create_cable()
        wire = cable.create_wire()
        wire.connect_pin(pin)
        instance = sdn.Instance()
        instance.reference = definition
        netlist.top_instance = instance
        href = next(sdn.get_hcables(cable))

        search = next(sdn.get_pins(href))
        self.assertEqual(search, pin)

    def test_of_netlist_and_library_and_port(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        port = definition.create_port()
        pin = port.create_pin()
        instance = sdn.Instance()
        instance.reference = definition

        search = next(netlist.get_pins())
        self.assertEqual(search, pin)

        search = next(sdn.get_pins(library))
        self.assertEqual(search, pin)

        search = next(sdn.get_pins(port))
        self.assertEqual(search, pin)
